#!/bin/bash
#
# Run select tests by setting ONLY, or as arguments to the script.
# Skip specific tests by setting EXCEPT.
#
set -e

ONLY=${ONLY:-"$*"}

LUSTRE=${LUSTRE:-$(dirname $0)/..}
. $LUSTRE/tests/test-framework.sh
init_test_env $@
init_logging

ALWAYS_EXCEPT="$SANITY_QUOTA_EXCEPT "
# Bug number for skipped test:
ALWAYS_EXCEPT+=""
# UPDATE THE COMMENT ABOVE WITH BUG NUMBERS WHEN CHANGING ALWAYS_EXCEPT!

# Test duration:                   30 min
[ "$SLOW" = "no" ] && EXCEPT_SLOW="61"

if [ "$mds1_FSTYPE" = "zfs" ]; then
	# bug number for skipped test: LU-6836
	ALWAYS_EXCEPT="$ALWAYS_EXCEPT  4a"

	# bug number:                        LU-2887
	# Test duration:                     21      9 min"
	[ "$SLOW" = "no" ] && EXCEPT_SLOW+=" 12a     9"
fi

build_test_filter

DIRECTIO=${DIRECTIO:-$LUSTRE/tests/directio}
ORIG_PWD=${PWD}
TSTID=${TSTID:-"$(id -u $TSTUSR)"}
TSTID2=${TSTID2:-"$(id -u $TSTUSR2)"}
TSTPRJID=${TSTPRJID:-1000}
BLK_SZ=1024
MAX_DQ_TIME=604800
MAX_IQ_TIME=604800
QTYPE="ugp"

require_dsh_mds || exit 0
require_dsh_ost || exit 0

# Does e2fsprogs support quota feature?
if [ $(facet_fstype $SINGLEMDS) == ldiskfs ] &&
	do_facet $SINGLEMDS "! $DEBUGFS -c -R supported_features |
		grep -q 'quota'"; then
	skip_env "e2fsprogs doesn't support quota" && exit 0
fi

QUOTALOG=${TESTSUITELOG:-$TMP/$(basename $0 .sh).log}

[ "$QUOTALOG" ] && rm -f $QUOTALOG || true

DIR=${DIR:-$MOUNT}
DIR2=${DIR2:-$MOUNT2}

QUOTA_AUTO_OLD=$QUOTA_AUTO
export QUOTA_AUTO=0

check_and_setup_lustre

ENABLE_PROJECT_QUOTAS=${ENABLE_PROJECT_QUOTAS:-true}

SHOW_QUOTA_USER="$LFS quota -v -u $TSTUSR $DIR"
SHOW_QUOTA_USERID="$LFS quota -v -u $TSTID $DIR"
SHOW_QUOTA_GROUP="$LFS quota -v -g $TSTUSR $DIR"
SHOW_QUOTA_GROUPID="$LFS quota -v -g $TSTID $DIR"
SHOW_QUOTA_PROJID="eval is_project_quota_supported && $LFS quota -v -p $TSTPRJID $DIR"
SHOW_QUOTA_INFO_USER="$LFS quota -t -u $DIR"
SHOW_QUOTA_INFO_GROUP="$LFS quota -t -g $DIR"
SHOW_QUOTA_INFO_PROJID="eval is_project_quota_supported && $LFS quota -t -p $DIR"

lustre_fail() {
	local fail_node=$1
	local fail_loc=$2
	local fail_val=${3:-0}
	local NODES=

	case $fail_node in
	mds_ost|mdt_ost) NODES="$(comma_list $(mdts_nodes) $(osts_nodes))";;
	mds|mdt) NODES="$(comma_list $(mdts_nodes))";;
	ost) NODES="$(comma_list $(osts_nodes))";;
	esac

	do_nodes $NODES "lctl set_param fail_val=$fail_val fail_loc=$fail_loc"
}

change_project()
{
	echo "lfs project $*"
	lfs project $* || error "lfs project $* failed"
}

RUNAS="runas -u $TSTID -g $TSTID"
RUNAS2="runas -u $TSTID2 -g $TSTID2"
DD="dd if=/dev/zero bs=1M"

FAIL_ON_ERROR=false

# clear quota limits for a user or a group
# usage: resetquota -u username
#        resetquota -g groupname
#	 resetquota -p projid

resetquota() {
	[ "$#" != 2 ] && error "resetquota: wrong number of arguments: $#"
	[ "$1" != "-u" -a "$1" != "-g" -a "$1" != "-p" ] &&
		error "resetquota: wrong specifier $1 passed"

	if [ $1 == "-p" ]; then
		is_project_quota_supported || return 0
	fi

	$LFS setquota "$1" "$2" -b 0 -B 0 -i 0 -I 0 $MOUNT ||
		error "clear quota for [type:$1 name:$2] failed"
	# give a chance to slave to release space
	sleep 1
}

quota_scan() {
	local local_ugp=$1
	local local_id=$2

	if [ "$local_ugp" == "a" -o "$local_ugp" == "u" ]; then
		$LFS quota -v -u $local_id $DIR
		log "Files for user ($local_id):"
		($LFS find --user $local_id $DIR | head -n 4 |
			xargs stat 2>/dev/null)
	fi

	if [ "$local_ugp" == "a" -o "$local_ugp" == "g" ]; then
		$LFS quota -v -g $local_id $DIR
		log "Files for group ($local_id):"
		($LFS find --group $local_id $DIR | head -n 4 |
			xargs stat 2>/dev/null)
	fi

	is_project_quota_supported || return 0
	if [ "$local_ugp" == "a" -o "$local_ugp" == "p" ]; then
		$LFS quota -v -p $TSTPRJID $DIR
		log "Files for project ($TSTPRJID):"
		($LFS find --projid $TSTPRJID $DIR | head -n 4 |
			xargs stat 2>/dev/null)
	fi
}

quota_error() {
	quota_scan $1 $2
	shift 2
	error "$*"
}

quota_log() {
	quota_scan $1 $2
	shift 2
	log "$*"
}

# get quota for a user or a group
# usage: getquota -u|-g|-p <username>|<groupname>|<projid> global|<obd_uuid> \
#		  bhardlimit|bsoftlimit|bgrace|ihardlimit|isoftlimit|igrace
getquota() {
	local spec
	local uuid

	sync_all_data > /dev/null 2>&1 || true

	[ "$#" != 4 ] && error "getquota: wrong number of arguments: $#"
	[ "$1" != "-u" -a "$1" != "-g" -a "$1" != "-p" ] &&
		error "getquota: wrong u/g/p specifier $1 passed"

	uuid="$3"

	case "$4" in
		curspace)   spec=1;;
		bsoftlimit) spec=2;;
		bhardlimit) spec=3;;
		bgrace)     spec=4;;
		curinodes)  spec=5;;
		isoftlimit) spec=6;;
		ihardlimit) spec=7;;
		igrace)     spec=8;;
		*)          error "unknown quota parameter $4";;
	esac

	[ "$uuid" = "global" ] && uuid=$DIR

	$LFS quota -v "$1" "$2" $DIR |
		awk 'BEGIN { num='$spec' } { if ($1 == "'$uuid'") \
		{ if (NF == 1) { getline } else { num++ } ; print $num;} }' \
		| tr -d "*"
}

# set mdt quota type
# usage: set_mdt_qtype ugp|u|g|p|none
set_mdt_qtype() {
	local qtype=$1
	local varsvc
	local mdts=$(get_facets MDS)
	local cmd
	[[ "$qtype" =~ "p" ]] && ! is_project_quota_supported &&
		qtype=$(tr -d 'p' <<<$qtype)

	if [[ $PERM_CMD = *"set_param -P"* ]]; then
		do_facet mgs $PERM_CMD \
			osd-*.$FSNAME-MDT*.quota_slave.enable=$qtype
	else
		do_facet mgs $PERM_CMD $FSNAME.quota.mdt=$qtype
	fi
	# we have to make sure each MDT received config changes
	for mdt in ${mdts//,/ }; do
		varsvc=${mdt}_svc
		cmd="$LCTL get_param -n "
		cmd=${cmd}osd-$(facet_fstype $mdt).${!varsvc}
		cmd=${cmd}.quota_slave.enabled

		if $(facet_up $mdt); then
			wait_update_facet $mdt "$cmd" "$qtype" || return 1
		fi
	done
	return 0
}

# set ost quota type
# usage: set_ost_qtype ugp|u|g|p|none
set_ost_qtype() {
	local qtype=$1
	local varsvc
	local osts=$(get_facets OST)
	local cmd
	[[ "$qtype" =~ "p" ]] && ! is_project_quota_supported &&
		qtype=$(tr -d 'p' <<<$qtype)

	if [[ $PERM_CMD = *"set_param -P"* ]]; then
		do_facet mgs $PERM_CMD \
			osd-*.$FSNAME-OST*.quota_slave.enable=$qtype
	else
		do_facet mgs $PERM_CMD $FSNAME.quota.ost=$qtype
	fi
	# we have to make sure each OST received config changes
	for ost in ${osts//,/ }; do
		varsvc=${ost}_svc
		cmd="$LCTL get_param -n "
		cmd=${cmd}osd-$(facet_fstype $ost).${!varsvc}
		cmd=${cmd}.quota_slave.enabled

		if $(facet_up $ost); then
			wait_update_facet $ost "$cmd" "$qtype" || return 1
		fi
	done
	return 0
}

wait_reintegration() {
	local ntype=$1
	local qtype=$2
	local max=$3
	local result="glb[1],slv[1],reint[0]"
	local varsvc
	local cmd
	local tgts

	if [ $ntype == "mdt" ]; then
		tgts=$(get_facets MDS)
	else
		tgts=$(get_facets OST)
	fi

	for tgt in ${tgts//,/ }; do
		varsvc=${tgt}_svc
		cmd="$LCTL get_param -n "
		cmd=${cmd}osd-$(facet_fstype $tgt).${!varsvc}
		cmd=${cmd}.quota_slave.info

		if $(facet_up $tgt); then
			wait_update_facet $tgt "$cmd |
				grep "$qtype" | awk '{ print \\\$3 }'" \
					"$result" $max || return 1
		fi
	done
	return 0
}

wait_mdt_reint() {
	local qtype=$1
	local max=${2:-90}

	if [[ "$qtype" =~ "u" ]]; then
		wait_reintegration "mdt" "user" $max || return 1
	fi

	if [[ "$qtype" =~ "g" ]]; then
		wait_reintegration "mdt" "group" $max || return 1
	fi

	if [[ "$qtype" =~ "p" ]]; then
		! is_project_quota_supported && return 0
		wait_reintegration "mdt" "project" $max || return 1
	fi
	return 0
}

wait_ost_reint() {
	local qtype=$1
	local max=${2:-90}

	if [[ "$qtype" =~ "u" ]]; then
		wait_reintegration "ost" "user" $max || return 1
	fi

	if [[ "$qtype" =~ "g" ]]; then
		wait_reintegration "ost" "group" $max || return 1
	fi

	if [[ "$qtype" =~ "p" ]]; then
		! is_project_quota_supported && return 0
		wait_reintegration "ost" "project" $max || return 1
	fi
	return 0
}

disable_project_quota() {
	is_project_quota_supported || return 0
	[ "$(facet_fstype $SINGLEMDS)" != "ldiskfs" ] && return 0
	stopall || error "failed to stopall (1)"

	for num in $(seq $MDSCOUNT); do
		do_facet mds$num $TUNE2FS -Q ^prj $(mdsdevname $num) ||
			error "tune2fs $(mdsdevname $num) failed"
	done

	for num in $(seq $OSTCOUNT); do
		do_facet ost$num $TUNE2FS -Q ^prj $(ostdevname $num) ||
			error "tune2fs $(ostdevname $num) failed"
	done

	mount
	setupall
}

wait_grace_time() {
	local qtype=$1
	local flavour=$2
	local extrasleep=${3:-5}
	local qarg

	case $qtype in
		u|g) qarg=$TSTUSR ;;
		p) qarg=$TSTPRJID ;;
		*) error "get_grace_time: Invalid quota type: $qtype"
	esac

	case $flavour in
		block)
			time=$(lfs quota -$qtype $qarg $DIR|
				   awk 'NR == 3{ print $5 }'| sed 's/s$//')
			;;
		file)
			time=$(lfs quota -$qtype $qarg $DIR|
				   awk 'NR == 3{ print $9 }'| sed 's/s$//')
			;;
		*)
			error "Unknown quota type: $flavour"
			;;
	esac

	echo "Sleep through grace ..."
	[ "$time" == "-" ] &&
	    error "Grace timeout was not set or quota not exceeded"
	if [ "$time" == "none" ]; then
	    echo "...Grace timeout already expired"
	else
		let time+=$extrasleep
		echo "...sleep $time seconds"
		sleep $time
	fi
}

setup_quota_test() {
	wait_delete_completed
	echo "Creating test directory"
	mkdir $DIR/$tdir || return 1
	chmod 0777 $DIR/$tdir || return 2
	# always clear fail_loc in case of fail_loc isn't cleared
	# properly when previous test failed
	lustre_fail mds_ost 0
	stack_trap cleanup_quota_test EXIT
}

cleanup_quota_test() {
	echo "Delete files..."
	rm -rf $DIR/$tdir
	[ -d $DIR/${tdir}_dom ] && rm -rf $DIR/${tdir}_dom
	echo "Wait for unlink objects finished..."
	wait_delete_completed
	sync_all_data || true
	reset_quota_settings || true
}

quota_show_check() {
	local bf=$1
	local ugp=$2
	local qid=$3
	local usage

	$LFS quota -v -$ugp $qid $DIR

	if [ "$bf" == "a" -o "$bf" == "b" ]; then
		usage=$(getquota -$ugp $qid global curspace)
		if [ -z $usage ]; then
			quota_error $ugp $qid \
				"Query block quota failed ($ugp:$qid)."
		else
			[ $usage -ne 0 ] && quota_log $ugp $qid \
				"Block quota isn't 0 ($ugp:$qid:$usage)."
		fi
	fi

	if [ "$bf" == "a" -o "$bf" == "f" ]; then
		usage=$(getquota -$ugp $qid global curinodes)
		if [ -z $usage ]; then
			quota_error $ugp $qid \
				"Query file quota failed ($ugp:$qid)."
		else
			[ $usage -ne 0 ] && quota_log $ugp $qid \
				"File quota isn't 0 ($ugp:$qid:$usage)."
		fi
	fi
}

enable_project_quota() {
	is_project_quota_supported || return 0
	[ "$(facet_fstype $SINGLEMDS)" != "ldiskfs" ] && return 0
	stopall || error "failed to stopall (1)"

	for num in $(seq $MDSCOUNT); do
		do_facet mds$num $TUNE2FS -O project $(mdsdevname $num) ||
			error "tune2fs $(mdsdevname $num) failed"
	done

	for num in $(seq $OSTCOUNT); do
		do_facet ost$num $TUNE2FS -O project $(ostdevname $num) ||
			error "tune2fs $(ostdevname $num) failed"
	done

	mount
	setupall
}
enable_project_quota

reset_quota_settings() {
	echo "Reset quota limits..."

	resetquota -u $TSTUSR
	resetquota -g $TSTUSR
	resetquota -u $TSTUSR2
	resetquota -g $TSTUSR2
	resetquota -p $TSTPRJID
}

get_quota_on_qsd() {
	local facet
	local device
	local spec
	local qid
	local qtype
	local output

	facet=$1
	device=$2
	case "$3" in
		usr) qtype="limit_user";;
		grp) qtype="limit_group";;
		prj) qtype="limit_project";;
		*)	   error "unknown quota parameter $3";;
	esac

	qid=$4
	case "$5" in
		hardlimit) spec=4;;
		softlimit) spec=6;;
		*)	   error "unknown quota parameter $5";;
	esac

	do_facet $facet $LCTL get_param osd-*.*-${device}.quota_slave.$qtype |
		awk '($3 == '$qid') {getline; print $'$spec'; exit;}' | tr -d ,
}

wait_quota_synced() {
	local value
	local facet=$1
	local device=$2
	local qtype=$3
	local qid=$4
	local limit_type=$5
	local limit_val=$6
	local interval=0

	value=$(get_quota_on_qsd $facet $device $qtype $qid $limit_type)
	while [[ $value != $limit_val ]]; do
		(( interval != 0 )) ||
			do_facet $facet $LCTL set_param \
				osd-*.*-${device}.quota_slave.force_reint=1

		echo $value
		(( interval <= 20 )) ||
			error "quota ($value) don't update on QSD, $limit_val"

		interval=$((interval + 1))
		sleep 1

		value=$(get_quota_on_qsd $facet $device $qtype $qid $limit_type)
	done
}

# enable quota debug
quota_init() {
	do_nodes $(comma_list $(nodes_list)) "lctl set_param debug=+quota"
}
quota_init
reset_quota_settings

check_runas_id_ret $TSTUSR $TSTUSR $RUNAS ||
	error "Please create user $TSTUSR($TSTID) and group $TSTUSR($TSTID)"
check_runas_id_ret $TSTUSR2 $TSTUSR2 $RUNAS2 ||
	error "Please create user $TSTUSR2($TSTID2) and group $TSTUSR2($TSTID2)"

test_quota_performance() {
	local TESTFILE="$DIR/$tdir/$tfile-0"
	local size=$1 # in MB
	local stime=$(date +%s)
	$RUNAS $DD of=$TESTFILE count=$size conv=fsync ||
		quota_error u $TSTUSR "write failure"
	local etime=$(date +%s)
	delta=$((etime - stime))
	if [ $delta -gt 0 ]; then
	    rate=$((size * 1024 / delta))
	    if [ $(facet_fstype $SINGLEMDS) = "zfs" ]; then
		# LU-2872 - see LU-2887 for fix
		[ $rate -gt 64 ] ||
			error "SLOW IO for $TSTUSR (user): $rate KB/sec"
	    else
		[ $rate -gt 1024 ] ||
			error "SLOW IO for $TSTUSR (user): $rate KB/sec"
	    fi
	fi
	rm -f $TESTFILE
}

# test basic quota performance b=21696
test_0() {
	local MB=100 # 100M
	[ "$SLOW" = "no" ] && MB=10

	local free_space=$(lfs_df | grep "summary" | awk '{print $4}')
	[ $free_space -le $((MB * 1024)) ] &&
		skip "not enough space ${free_space} KB, " \
			"required $((MB * 1024)) KB"
	setup_quota_test || error "setup quota failed with $?"

	set_ost_qtype "none" || error "disable ost quota failed"
	test_quota_performance $MB

	set_ost_qtype $QTYPE || error "enable ost quota failed"
	$LFS setquota -u $TSTUSR -b 0 -B 10G -i 0 -I 0 $DIR ||
		error "set quota failed"
	test_quota_performance $MB

}
run_test 0 "Test basic quota performance"

check_write_fallocate() {
	local testfile="$1"
	local qtype="$2"
	local limit=$3
	local short_qtype=${qtype:0:1}

	count=$((limit/2))
	log "Write ${count}M Using Fallocate"
	$RUNAS fallocate -l${count}M $testfile ||
		quota_error $short_qtype $TSTUSR "Write ${count}M fail"

	cancel_lru_locks osc
	sync; sync_all_data || true
	sleep 2

	count=$((limit + 1))
	log "Write ${count}M Using Fallocate"
	$RUNAS fallocate -l${count}M $testfile &&
		quota_error $short_qtype $TSTUSR \
		"Write success, expect EDQUOT" || true
}

# test block hardlimit
test_1() {
	local LIMIT=10  # 10M
	local TESTFILE="$DIR/$tdir/$tfile-0"

	setup_quota_test || error "setup quota failed with $?"

	# enable ost quota
	set_ost_qtype $QTYPE || error "enable ost quota failed"

	# test for user
	log "User quota (block hardlimit:$LIMIT MB)"
	$LFS setquota -u $TSTUSR -b 0 -B ${LIMIT}M -i 0 -I 0 $DIR ||
		error "set user quota failed"

	# make sure the system is clean
	local USED=$(getquota -u $TSTUSR global curspace)
	[ $USED -ne 0 ] && error "Used space($USED) for user $TSTUSR isn't 0."

	$SETSTRIPE $TESTFILE -c 1 || error "setstripe $TESTFILE failed"
	chown $TSTUSR.$TSTUSR $TESTFILE || error "chown $TESTFILE failed"

	log "Write..."
	$RUNAS $DD of=$TESTFILE count=$((LIMIT/2)) ||
		quota_error u $TSTUSR "user write failure, but expect success"
	log "Write out of block quota ..."
	# this time maybe cache write, ignore it's failure
	$RUNAS $DD of=$TESTFILE count=$((LIMIT/2)) seek=$((LIMIT/2)) || true
	# flush cache, ensure noquota flag is set on client
	cancel_lru_locks osc
	sync; sync_all_data || true
	$RUNAS $DD of=$TESTFILE count=1 seek=$LIMIT &&
		quota_error u $TSTUSR "user write success, but expect EDQUOT"

	rm -f $TESTFILE
	wait_delete_completed || error "wait_delete_completed failed"
	sync_all_data || true
	USED=$(getquota -u $TSTUSR global curspace)
	[ $USED -ne 0 ] && quota_error u $TSTUSR \
		"user quota isn't released after deletion"
	resetquota -u $TSTUSR

	# test for group
	log "--------------------------------------"
	log "Group quota (block hardlimit:$LIMIT MB)"
	$LFS setquota -g $TSTUSR -b 0 -B ${LIMIT}M -i 0 -I 0 $DIR ||
		error "set group quota failed"

	TESTFILE="$DIR/$tdir/$tfile-1"
	# make sure the system is clean
	USED=$(getquota -g $TSTUSR global curspace)
	[ $USED -ne 0 ] && error "Used space ($USED) for group $TSTUSR isn't 0"

	$SETSTRIPE $TESTFILE -c 1 || error "setstripe $TESTFILE failed"
	chown $TSTUSR.$TSTUSR $TESTFILE || error "chown $TESTFILE failed"

	log "Write ..."
	$RUNAS $DD of=$TESTFILE count=$((LIMIT/2)) ||
		quota_error g $TSTUSR "Group write failure, but expect success"
	log "Write out of block quota ..."
	# this time maybe cache write, ignore it's failure
	$RUNAS $DD of=$TESTFILE count=$((LIMIT/2)) seek=$((LIMIT/2)) || true
	cancel_lru_locks osc
	sync; sync_all_data || true
	$RUNAS $DD of=$TESTFILE count=10 seek=$LIMIT &&
		quota_error g $TSTUSR "Group write success, but expect EDQUOT"
	rm -f $TESTFILE
	wait_delete_completed || error "wait_delete_completed failed"
	sync_all_data || true
	USED=$(getquota -g $TSTUSR global curspace)
	[ $USED -ne 0 ] && quota_error g $TSTUSR \
				"Group quota isn't released after deletion"
	resetquota -g $TSTUSR

	if ! is_project_quota_supported; then
		echo "Project quota is not supported"
		return 0
	fi

	TESTFILE="$DIR/$tdir/$tfile-2"
	# make sure the system is clean
	USED=$(getquota -p $TSTPRJID global curspace)
	[ $USED -ne 0 ] &&
		error "used space($USED) for project $TSTPRJID isn't 0"

	# test for Project
	log "--------------------------------------"
	log "Project quota (block hardlimit:$LIMIT mb)"
	$LFS setquota -p $TSTPRJID -b 0 -B ${LIMIT}M -i 0 -I 0 $DIR ||
		error "set project quota failed"

	$SETSTRIPE $TESTFILE -c 1 || error "setstripe $TESTFILE failed"
	chown $TSTUSR:$TSTUSR $TESTFILE || error "chown $TESTFILE failed"
	change_project -p $TSTPRJID $TESTFILE

	log "write ..."
	$RUNAS $DD of=$TESTFILE count=$((LIMIT/2)) || quota_error p $TSTPRJID \
		"project write failure, but expect success"
	log "write out of block quota ..."
	# this time maybe cache write, ignore it's failure
	$RUNAS $DD of=$TESTFILE count=$((LIMIT/2)) seek=$((LIMIT/2)) || true
	cancel_lru_locks osc
	sync; sync_all_data || true
	$RUNAS $DD of=$TESTFILE count=10 seek=$LIMIT && quota_error p \
		$TSTPRJID "project write success, but expect EDQUOT"

	# cleanup
	cleanup_quota_test

	USED=$(getquota -p $TSTPRJID global curspace)
	[ $USED -ne 0 ] && quota_error p $TSTPRJID \
		"project quota isn't released after deletion"

	resetquota -p $TSTPRJID
}
run_test 1 "Block hard limit (normal use and out of quota)"

test_1h() {
	local limit=10  # 10M
	local testfile="$DIR/$tdir/$tfile-0"

	check_set_fallocate_or_skip

	setup_quota_test || error "setup quota failed with $?"

	# enable ost quota
	set_ost_qtype $QTYPE || error "enable ost quota failed"

	# test for user
	log "User quota (block hardlimit:$limit MB)"
	$LFS setquota -u $TSTUSR -b 0 -B ${limit}M -i 0 -I 0 $DIR ||
		error "set user quota failed"

	# make sure the system is clean
	local used=$(getquota -u $TSTUSR global curspace)
	[ $used -ne 0 ] && error "Used space($used) for user $TSTUSR isn't 0."

	$LFS setstripe $testfile -i 0 -c 1 || error "setstripe $testfile failed"
	chown $TSTUSR.$TSTUSR $testfile || error "chown $testfile failed"

	wait_quota_synced ost1 OST0000 usr $TSTID hardlimit $((limit*1024))

	check_write_fallocate $testfile "user" $limit

	rm -f $testfile
	wait_delete_completed || error "wait_delete_completed failed"
	sync_all_data || true
	used=$(getquota -u $TSTUSR global curspace)
	[ $used -eq 0 ] || quota_error u $TSTUSR \
		"user quota isn't released after deletion"
}
run_test 1h "Block hard limit test using fallocate"

test_1j() {
	local limit=20 # MB
	local testfile="$DIR/$tdir/$tfile-0"

	(( $OST1_VERSION >= $(version_code 2.12.9.11) )) ||
		skip "need OST >= 2.12.9.11 for root project quota"

	is_project_quota_supported ||
		skip "skip project quota unsupported"

	setup_quota_test || error "setup quota failed with $?"

	# enable ost quota
	set_ost_qtype $QTYPE || error "enable ost quota failed"

	# test for Project
	log "--------------------------------------"
	log "Project quota (block hardlimit:$limit mb)"
	$LFS setquota -p $TSTPRJID -b 0 -B ${limit}M -i 0 -I 0 $DIR ||
		error "set project quota failed"

	$LFS setstripe $testfile -c 1 -i 0 || error "setstripe $testfile failed"
	change_project -p $TSTPRJID $testfile

	local procf=osd-$ost1_FSTYPE.$FSNAME-OST0000.quota_slave.root_prj_enable
	do_facet ost1 $LCTL set_param $procf=1 ||
		error "enable root quotas for project failed"
	stack_trap "do_facet ost1 $LCTL set_param $procf=0"

	runas -u 0 -g 0 $DD of=$testfile count=$limit oflag=direct || true
	runas -u 0 -g 0 $DD of=$testfile count=$((limit/2)) seek=$limit oflag=direct &&
		quota_error "project" $TSTPRJID "root write to project success"

	do_facet ost1 $LCTL set_param $procf=0 ||
		error "disable root quotas for project failed"

	runas -u 0 -g 0 $DD of=$testfile count=$limit seek=$limit oflag=direct ||
		quota_error "project" $TSTPRJID "root write to project failed"

	# cleanup
	cleanup_quota_test

	used=$(getquota -p $TSTPRJID global curspace)
	(( $used == 0 )) || quota_error p $TSTPRJID \
		"project quota isn't released after deletion"
}
run_test 1j "Enable project quota enforcement for root"

# test inode hardlimit
test_2() {
	local TESTFILE="$DIR/$tdir/$tfile-0"
	local LIMIT=$(do_facet mds1 $LCTL get_param -n \
		qmt.$FSNAME-QMT0000.md-0x0.info |
		awk '/least qunit/{ print $3 }')
	local L2=$(do_facet mds1 $LCTL get_param -n \
		qmt.$FSNAME-QMT0000.md-0x0.soft_least_qunit)

	[ $L2 -le $LIMIT ] || LIMIT=$L2

	[ "$SLOW" = "no" ] || LIMIT=$((LIMIT * 1024))

	local FREE_INODES=$(mdt_free_inodes 0)
	echo "$FREE_INODES free inodes on master MDT"
	[ $FREE_INODES -lt $LIMIT ] &&
		skip "not enough free inodes $FREE_INODES required $LIMIT"

	setup_quota_test || error "setup quota failed with $?"

	# enable mdt quota
	set_mdt_qtype $QTYPE || error "enable mdt quota failed"

	# test for user
	log "User quota (inode hardlimit:$LIMIT files)"
	$LFS setquota -u $TSTUSR -b 0 -B 0 -i 0 -I $LIMIT $DIR ||
		error "set user quota failed"

	# make sure the system is clean
	local USED=$(getquota -u $TSTUSR global curinodes)
	[ $USED -ne 0 ] && error "Used inodes($USED) for user $TSTUSR isn't 0."

	log "Create $LIMIT files ..."
	$RUNAS createmany -m ${TESTFILE} $LIMIT ||
		quota_error u $TSTUSR "user create failure, but expect success"
	log "Create out of file quota ..."
	$RUNAS touch ${TESTFILE}_xxx &&
		quota_error u $TSTUSR "user create success, but expect EDQUOT"

	# cleanup
	unlinkmany ${TESTFILE} $LIMIT || error "unlinkmany $TESTFILE failed"
	rm -f ${TESTFILE}_xxx
	wait_delete_completed

	USED=$(getquota -u $TSTUSR global curinodes)
	[ $USED -ne 0 ] && quota_error u $TSTUSR \
		"user quota isn't released after deletion"
	resetquota -u $TSTUSR

	# test for group
	log "--------------------------------------"
	log "Group quota (inode hardlimit:$LIMIT files)"
	$LFS setquota -g $TSTUSR -b 0 -B 0 -i 0 -I $LIMIT $DIR ||
		error "set group quota failed"

	TESTFILE=$DIR/$tdir/$tfile-1
	# make sure the system is clean
	USED=$(getquota -g $TSTUSR global curinodes)
	[ $USED -ne 0 ] && error "Used inodes($USED) for group $TSTUSR isn't 0."

	log "Create $LIMIT files ..."
	$RUNAS createmany -m ${TESTFILE} $LIMIT ||
		quota_error g $TSTUSR "group create failure, but expect success"
	log "Create out of file quota ..."
	$RUNAS touch ${TESTFILE}_xxx &&
		quota_error g $TSTUSR "group create success, but expect EDQUOT"

	# cleanup
	unlinkmany ${TESTFILE} $LIMIT || error "unlinkmany $TESTFILE failed"
	rm -f ${TESTFILE}_xxx
	wait_delete_completed

	USED=$(getquota -g $TSTUSR global curinodes)
	[ $USED -ne 0 ] && quota_error g $TSTUSR \
		"user quota isn't released after deletion"

	resetquota -g $TSTUSR
	! is_project_quota_supported &&
		echo "Skip project quota is not supported" && return 0

	# test for project
	log "--------------------------------------"
	log "Project quota (inode hardlimit:$LIMIT files)"
	$LFS setquota -p $TSTPRJID -b 0 -B 0 -i 0 -I $LIMIT $DIR ||
		error "set project quota failed"

	TESTFILE=$DIR/$tdir/$tfile-1
	# make sure the system is clean
	USED=$(getquota -p $TSTPRJID global curinodes)
	[ $USED -ne 0 ] &&
		error "Used inodes($USED) for project $TSTPRJID isn't 0"

	change_project -sp $TSTPRJID $DIR/$tdir
	log "Create $LIMIT files ..."
	$RUNAS createmany -m ${TESTFILE} $((LIMIT-1)) || quota_error p \
		$TSTPRJID "project create fail, but expect success"
	log "Create out of file quota ..."
	$RUNAS touch ${TESTFILE}_xxx && quota_error p $TSTPRJID \
		"project create success, but expect EDQUOT"
	change_project -C $DIR/$tdir

	cleanup_quota_test
	USED=$(getquota -p $TSTPRJID global curinodes)
	[ $USED -eq 0 ] || quota_error p $TSTPRJID \
		"project quota isn't released after deletion"

}
run_test 2 "File hard limit (normal use and out of quota)"

test_block_soft() {
	local TESTFILE=$1
	local GRACE=$2
	local LIMIT=$3
	local OFFSET=0
	local qtype=$4
	local soft_limit=$(do_facet $SINGLEMDS $LCTL get_param -n \
		qmt.$FSNAME-QMT0000.dt-0x0.soft_least_qunit)

	setup_quota_test

	$SETSTRIPE $TESTFILE -c 1 -i 0
	chown $TSTUSR.$TSTUSR $TESTFILE
	[ "$qtype" == "p" ] && is_project_quota_supported &&
		change_project -p $TSTPRJID $TESTFILE

	echo "Write up to soft limit"
	$RUNAS $DD of=$TESTFILE count=$LIMIT ||
		quota_error a $TSTUSR "write failure, but expect success"
	OFFSET=$((LIMIT * 1024))
	cancel_lru_locks osc

	echo "Write to exceed soft limit"
	$RUNAS dd if=/dev/zero of=$TESTFILE bs=1K count=10 seek=$OFFSET ||
		quota_error a $TSTUSR "write failure, but expect success"
	OFFSET=$((OFFSET + 1024)) # make sure we don't write to same block
	cancel_lru_locks osc

	echo "mmap write when over soft limit"
	$RUNAS $MULTIOP $TESTFILE.mmap OT40960SMW ||
		quota_error a $TSTUSR "mmap write failure, but expect success"
	cancel_lru_locks osc

	$SHOW_QUOTA_USER
	$SHOW_QUOTA_GROUP
	$SHOW_QUOTA_PROJID
	$SHOW_QUOTA_INFO_USER
	$SHOW_QUOTA_INFO_GROUP
	$SHOW_QUOTA_INFO_PROJID

	echo "Write before timer goes off"
	$RUNAS dd if=/dev/zero of=$TESTFILE bs=1K count=10 seek=$OFFSET ||
		quota_error a $TSTUSR "write failure, but expect success"
	OFFSET=$((OFFSET + 1024))
	cancel_lru_locks osc

	wait_grace_time $qtype "block"

	$SHOW_QUOTA_USER
	$SHOW_QUOTA_GROUP
	$SHOW_QUOTA_PROJID
	$SHOW_QUOTA_INFO_USER
	$SHOW_QUOTA_INFO_GROUP
	$SHOW_QUOTA_INFO_PROJID

	echo "Write after timer goes off"
	# maybe cache write, ignore.
	# write up to soft least quint to consume all
	# possible slave granted space.
	$RUNAS dd if=/dev/zero of=$TESTFILE bs=1K \
		count=$soft_limit seek=$OFFSET || true
	OFFSET=$((OFFSET + soft_limit))
	cancel_lru_locks osc
	$RUNAS dd if=/dev/zero of=$TESTFILE bs=1K count=10 seek=$OFFSET &&
		quota_error a $TSTUSR "write success, but expect EDQUOT"

	$SHOW_QUOTA_USER
	$SHOW_QUOTA_GROUP
	$SHOW_QUOTA_PROJID
	$SHOW_QUOTA_INFO_USER
	$SHOW_QUOTA_INFO_GROUP
	$SHOW_QUOTA_INFO_PROJID

	echo "Unlink file to stop timer"
	rm -f $TESTFILE
	wait_delete_completed
	sync_all_data || true

	$SHOW_QUOTA_USER
	$SHOW_QUOTA_GROUP
	$SHOW_QUOTA_PROJID
	$SHOW_QUOTA_INFO_USER
	$SHOW_QUOTA_INFO_GROUP
	$SHOW_QUOTA_INFO_PROJID

	$SETSTRIPE $TESTFILE -c 1 -i 0
	chown $TSTUSR.$TSTUSR $TESTFILE
	[ "$qtype" == "p" ] && change_project -p $TSTPRJID $TESTFILE

	echo "Write ..."
	$RUNAS $DD of=$TESTFILE count=$LIMIT ||
		quota_error a $TSTUSR "write failure, but expect success"
	# cleanup
	cleanup_quota_test
}

# block soft limit
test_3() {
	local GRACE=20 # 20s
	if [ $(facet_fstype $SINGLEMDS) = "zfs" ]; then
	    GRACE=60
	fi
	local TESTFILE=$DIR/$tdir/$tfile-0

	# get minimum soft qunit size
	local LIMIT=$(( $(do_facet $SINGLEMDS $LCTL get_param -n \
		qmt.$FSNAME-QMT0000.dt-0x0.soft_least_qunit) / 1024 ))

	set_ost_qtype $QTYPE || error "enable ost quota failed"

	echo "User quota (soft limit:$LIMIT MB  grace:$GRACE seconds)"
	# make sure the system is clean
	local USED=$(getquota -u $TSTUSR global curspace)
	[ $USED -ne 0 ] && error "Used space($USED) for user $TSTUSR isn't 0."

	$LFS setquota -t -u --block-grace $GRACE --inode-grace \
		$MAX_IQ_TIME $DIR || error "set user grace time failed"
	$LFS setquota -u $TSTUSR -b ${LIMIT}M -B 0 -i 0 -I 0 $DIR ||
		error "set user quota failed"

	test_block_soft $TESTFILE $GRACE $LIMIT "u"
	resetquota -u $TSTUSR

	echo "Group quota (soft limit:$LIMIT MB  grace:$GRACE seconds)"
	TESTFILE=$DIR/$tdir/$tfile-1
	# make sure the system is clean
	USED=$(getquota -g $TSTUSR global curspace)
	[ $USED -ne 0 ] && error "Used space($USED) for group $TSTUSR isn't 0."

	$LFS setquota -t -g --block-grace $GRACE --inode-grace \
		$MAX_IQ_TIME $DIR || error "set group grace time failed"
	$LFS setquota -g $TSTUSR -b ${LIMIT}M -B 0 -i 0 -I 0 $DIR ||
		error "set group quota failed"

	test_block_soft $TESTFILE $GRACE $LIMIT "g"
	resetquota -g $TSTUSR

	if is_project_quota_supported; then
		echo "Project quota (soft limit:$LIMIT MB  grace:$GRACE sec)"
		TESTFILE=$DIR/$tdir/$tfile-2
		# make sure the system is clean
		USED=$(getquota -p $TSTPRJID global curspace)
		[ $USED -ne 0 ] && error \
			"Used space($USED) for project $TSTPRJID isn't 0."

		$LFS setquota -t -p --block-grace $GRACE --inode-grace \
			$MAX_IQ_TIME $DIR ||
				error "set project grace time failed"
		$LFS setquota -p $TSTPRJID -b ${LIMIT}M -B 0 -i 0 -I 0 \
			$DIR || error "set project quota failed"

		test_block_soft $TESTFILE $GRACE $LIMIT "p"
		resetquota -p $TSTPRJID
		$LFS setquota -t -p --block-grace $MAX_DQ_TIME --inode-grace \
			$MAX_IQ_TIME $DIR ||
				error "restore project grace time failed"
	fi

	# cleanup
	$LFS setquota -t -u --block-grace $MAX_DQ_TIME --inode-grace \
		$MAX_IQ_TIME $DIR || error "restore user grace time failed"
	$LFS setquota -t -g --block-grace $MAX_DQ_TIME --inode-grace \
		$MAX_IQ_TIME $DIR || error "restore group grace time failed"
}
run_test 3 "Block soft limit (start timer, timer goes off, stop timer)"

test_file_soft() {
	local TESTFILE=$1
	local LIMIT=$2
	local grace=$3
	local qtype=$4

	setup_quota_test
	is_project_quota_supported && change_project -sp $TSTPRJID $DIR/$tdir

	echo "Create files to exceed soft limit"
	$RUNAS createmany -m ${TESTFILE}_ $((LIMIT + 1)) ||
		quota_error a $TSTUSR "create failure, but expect success"
	local trigger_time=$(date +%s)

	sync_all_data || true

	local cur_time=$(date +%s)
	[ $(($cur_time - $trigger_time)) -ge $grace ] &&
		error "Passed grace time $grace, $trigger_time, $cur_time"

	echo "Create file before timer goes off"
	$RUNAS touch ${TESTFILE}_before ||
		quota_error a $TSTUSR "failed create before timer expired," \
			"but expect success. $trigger_time, $cur_time"
	sync_all_data || true

	wait_grace_time $qtype "file"

	$SHOW_QUOTA_USER
	$SHOW_QUOTA_GROUP
	$SHOW_QUOTA_PROJID
	$SHOW_QUOTA_INFO_USER
	$SHOW_QUOTA_INFO_GROUP
	$SHOW_QUOTA_INFO_PROJID

	echo "Create file after timer goes off"
	# There is a window that space is accounted in the quota usage but
	# hasn't been decreased from the pending write, if we acquire quota
	# in this window, we'll acquire more than we needed.
	$RUNAS touch ${TESTFILE}_after_1 ${TESTFILE}_after_2 || true
	sync_all_data || true
	$RUNAS touch ${TESTFILE}_after_3 &&
		quota_error a $TSTUSR "create after timer expired," \
			"but expect EDQUOT"
	sync_all_data || true

	$SHOW_QUOTA_USER
	$SHOW_QUOTA_GROUP
	$SHOW_QUOTA_PROJID
	$SHOW_QUOTA_INFO_USER
	$SHOW_QUOTA_INFO_GROUP
	$SHOW_QUOTA_INFO_PROJID

	echo "Unlink files to stop timer"
	find $(dirname $TESTFILE) -name "$(basename ${TESTFILE})*" | xargs rm -f
	wait_delete_completed

	echo "Create file"
	$RUNAS touch ${TESTFILE}_xxx ||
		quota_error a $TSTUSR "touch after timer stop failure," \
			"but expect success"
	sync_all_data || true

	# cleanup
	cleanup_quota_test
}

# file soft limit
test_4a() {
	local LIMIT=$(do_facet $SINGLEMDS $LCTL get_param -n \
		qmt.$FSNAME-QMT0000.md-0x0.soft_least_qunit)
	local TESTFILE=$DIR/$tdir/$tfile-0
	local GRACE=12

	set_mdt_qtype $QTYPE || error "enable mdt quota failed"

	echo "User quota (soft limit:$LIMIT files  grace:$GRACE seconds)"
	# make sure the system is clean
	local USED=$(getquota -u $TSTUSR global curinodes)
	[ $USED -ne 0 ] && error "Used space($USED) for user $TSTUSR isn't 0."

	$LFS setquota -t -u --block-grace $MAX_DQ_TIME --inode-grace \
		$GRACE $DIR || error "set user grace time failed"
	$LFS setquota -u $TSTUSR -b 0 -B 0 -i $LIMIT -I 0 $DIR ||
		error "set user quota failed"

	test_file_soft $TESTFILE $LIMIT $GRACE "u"
	resetquota -u $TSTUSR

	echo "Group quota (soft limit:$LIMIT files  grace:$GRACE seconds)"
	# make sure the system is clean
	USED=$(getquota -g $TSTUSR global curinodes)
	[ $USED -ne 0 ] && error "Used space($USED) for group $TSTUSR isn't 0."

	$LFS setquota -t -g --block-grace $MAX_DQ_TIME --inode-grace \
		$GRACE $DIR || error "set group grace time failed"
	$LFS setquota -g $TSTUSR -b 0 -B 0 -i $LIMIT -I 0 $DIR ||
		error "set group quota failed"
	TESTFILE=$DIR/$tdir/$tfile-1

	test_file_soft $TESTFILE $LIMIT $GRACE "g"
	resetquota -g $TSTUSR

	if is_project_quota_supported; then
		echo "Project quota (soft limit:$LIMIT files grace:$GRACE sec)"
		# make sure the system is clean
		USED=$(getquota -p $TSTPRJID global curinodes)
		[ $USED -ne 0 ] && error \
			"Used space($USED) for project $TSTPRJID isn't 0."

		$LFS setquota -t -p --block-grace $MAX_DQ_TIME --inode-grace \
			$GRACE $DIR || error "set project grace time failed"
		$LFS setquota -p $TSTPRJID -b 0 -B 0 -i $LIMIT -I 0 $DIR ||
			error "set project quota failed"

		TESTFILE=$DIR/$tdir/$tfile-1
		# one less than limit, because of parent directory included.
		test_file_soft $TESTFILE $((LIMIT-1)) $GRACE "p"
		resetquota -p $TSTPRJID
		$LFS setquota -t -p --block-grace $MAX_DQ_TIME --inode-grace \
			$MAX_IQ_TIME $DIR ||
				error "restore project grace time failed"
	fi

	# cleanup
	$LFS setquota -t -u --block-grace $MAX_DQ_TIME --inode-grace \
		$MAX_IQ_TIME $DIR || error "restore user grace time failed"
	$LFS setquota -t -g --block-grace $MAX_DQ_TIME --inode-grace \
		$MAX_IQ_TIME $DIR || error "restore group grace time failed"
}
run_test 4a "File soft limit (start timer, timer goes off, stop timer)"

test_4b() {
	local GR_STR1="1w3d"
	local GR_STR2="1000s"
	local GR_STR3="5s"
	local GR_STR4="1w2d3h4m5s"
	local GR_STR5="5c"
	local GR_STR6="18446744073709551615"
	local GR_STR7="-1"

	wait_delete_completed

	# test of valid grace strings handling
	echo "Valid grace strings test"
	$LFS setquota -t -u --block-grace $GR_STR1 --inode-grace \
		$GR_STR2 $DIR || error "set user grace time failed"
	$LFS quota -u -t $DIR | grep "Block grace time: $GR_STR1"
	$LFS setquota -t -g --block-grace $GR_STR3 --inode-grace \
		$GR_STR4 $DIR || error "set group grace time quota failed"
	$LFS quota -g -t $DIR | grep "Inode grace time: $GR_STR4"

	# test of invalid grace strings handling
	echo "  Invalid grace strings test"
	! $LFS setquota -t -u --block-grace $GR_STR4 --inode-grace $GR_STR5 $DIR
	! $LFS setquota -t -g --block-grace $GR_STR4 --inode-grace $GR_STR6 $DIR
	! $LFS setquota -t -g --block-grace $GR_STR4 --inode-grace \
		$GR_STR7 $DIR

	# cleanup
	$LFS setquota -t -u --block-grace $MAX_DQ_TIME --inode-grace \
		$MAX_IQ_TIME $DIR || error "restore user grace time failed"
	$LFS setquota -t -g --block-grace $MAX_DQ_TIME --inode-grace \
		$MAX_IQ_TIME $DIR || error "restore group grace time failed"
}
run_test 4b "Grace time strings handling"

# chown & chgrp (chown & chgrp successfully even out of block/file quota)
test_5() {
	local BLIMIT=10 # 10M
	local ILIMIT=10 # 10 inodes

	setup_quota_test || error "setup quota failed with $?"

	set_mdt_qtype $QTYPE || error "enable mdt quota failed"
	set_ost_qtype $QTYPE || error "enable ost quota failed"

	echo "Set quota limit (0 ${BLIMIT}M 0 $ILIMIT) for $TSTUSR.$TSTUSR"
	$LFS setquota -u $TSTUSR -b 0 -B ${BLIMIT}M -i 0 -I $ILIMIT $DIR ||
		error "set user quota failed"
	$LFS setquota -g $TSTUSR -b 0 -B ${BLIMIT}M -i 0 -I $ILIMIT $DIR ||
	if is_project_quota_supported; then
		error "set group quota failed"
		$LFS setquota -p $TSTPRJID -b 0 -B ${BLIMIT}M -i 0 \
			-I $ILIMIT $DIR || error "set project quota failed"
	fi

	# make sure the system is clean
	local USED=$(getquota -u $TSTUSR global curinodes)
	[ $USED -ne 0 ] && error "Used inode($USED) for user $TSTUSR isn't 0."
	USED=$(getquota -g $TSTUSR global curinodes)
	[ $USED -ne 0 ] && error "Used inode($USED) for group $TSTUSR isn't 0."
	USED=$(getquota -u $TSTUSR global curspace)
	[ $USED -ne 0 ] && error "Used block($USED) for user $TSTUSR isn't 0."
	USED=$(getquota -g $TSTUSR global curspace)
	[ $USED -ne 0 ] && error "Used block($USED) for group $TSTUSR isn't 0."
	if is_project_quota_supported; then
		USED=$(getquota -p $TSTPRJID global curinodes)
		[ $USED -ne 0 ] &&
			error "Used inode($USED) for project $TSTPRJID isn't 0."
		USED=$(getquota -p $TSTPRJID global curspace)
		[ $USED -ne 0 ] &&
			error "Used block($USED) for project $TSTPRJID isn't 0."
	fi

	echo "Create more than $ILIMIT files and more than $BLIMIT MB ..."
	createmany -m $DIR/$tdir/$tfile-0_ $((ILIMIT + 1)) ||
		error "create failure, expect success"
	if is_project_quota_supported; then
		touch $DIR/$tdir/$tfile-0_1
		change_project -p $TSTPRJID $DIR/$tdir/$tfile-0_1
	fi
	$DD of=$DIR/$tdir/$tfile-0_1 count=$((BLIMIT+1)) ||
		error "write failure, expect success"

	echo "Chown files to $TSTUSR.$TSTUSR ..."
	for i in $(seq 0 $ILIMIT); do
		chown $TSTUSR.$TSTUSR $DIR/$tdir/$tfile-0_$i ||
			quota_error a $TSTUSR "chown failure, expect success"
	done

	# cleanup
	unlinkmany $DIR/$tdir/$tfile-0_ $((ILIMIT + 1)) ||
		error "unlinkmany $DIR/$tdir/$tfile-0_ failed"

}
run_test 5 "Chown & chgrp successfully even out of block/file quota"

# test dropping acquire request on master
test_6() {
	local LIMIT=3 # 3M

	# Clear dmesg so watchdog is not triggered by previous
	# test output
	do_facet ost1 dmesg -c > /dev/null

	setup_quota_test || error "setup quota failed with $?"

	# make sure the system is clean
	local USED=$(getquota -u $TSTUSR global curspace)
	[ $USED -ne 0 ] && error "Used space($USED) for user $TSTUSR isn't 0."

	# make sure no granted quota on ost
	set_ost_qtype $QTYPE || error "enable ost quota failed"
	resetquota -u $TSTUSR

	# create file for $TSTUSR
	local TESTFILE=$DIR/$tdir/$tfile-$TSTUSR
	$SETSTRIPE $TESTFILE -c 1 -i 0 || error "setstripe $TESTFILE failed"
	chown $TSTUSR.$TSTUSR $TESTFILE || error "chown $TESTFILE failed"

	# create file for $TSTUSR2
	local TESTFILE2=$DIR/$tdir/$tfile-$TSTUSR2
	$SETSTRIPE $TESTFILE2 -c 1 -i 0 || error "setstripe $TESTFILE2 failed"
	chown $TSTUSR2.$TSTUSR2 $TESTFILE2 || error "chown $TESTFILE2 failed"

	# cache per-ID lock for $TSTUSR on slave
	$LFS setquota -u $TSTUSR -b 0 -B ${LIMIT}M -i 0 -I 0 $DIR ||
		error "set quota failed"
	$RUNAS $DD of=$TESTFILE count=1 ||
		error "write $TESTFILE failure, expect success"
	$RUNAS2 $DD of=$TESTFILE2 count=1 ||
		error "write $TESTFILE2 failure, expect success"

	if at_is_enabled; then
		at_max_saved=$(at_max_get ost1)
		at_max_set $TIMEOUT ost1

		# write to enforced ID ($TSTUSR) to exceed limit to make sure
		# DQACQ is sent, which makes at_max to take effect
		$RUNAS $DD of=$TESTFILE count=$LIMIT seek=1 oflag=sync \
								conv=notrunc
		rm -f $TESTFILE
		wait_delete_completed
	fi

	sync; sync
	sync_all_data || true

	#define QUOTA_DQACQ 601
	#define OBD_FAIL_PTLRPC_DROP_REQ_OPC 0x513
	lustre_fail mds 0x513 601

	do_facet ost1 $LCTL set_param \
			osd-*.$FSNAME-OST*.quota_slave.timeout=$((TIMEOUT / 2))

	# write to un-enforced ID ($TSTUSR2) should succeed
	$RUNAS2 $DD of=$TESTFILE2 count=$LIMIT seek=1 oflag=sync conv=notrunc ||
		error "write failure, expect success"

	# write to enforced ID ($TSTUSR) in background, exceeding limit
	# to make sure DQACQ is sent
	$RUNAS $DD of=$TESTFILE count=$LIMIT seek=1 oflag=sync conv=notrunc &
	DDPID=$!

	# watchdog timer uses a factor of 2
	echo "Sleep for $((TIMEOUT * 2 + 1)) seconds ..."
	sleep $((TIMEOUT * 2 + 1))

	[ $at_max_saved -ne 0 ] && at_max_set $at_max_saved ost1

	# write should be blocked and never finished
	if ! ps -p $DDPID  > /dev/null 2>&1; then
		lustre_fail mds 0 0
		error "write finished incorrectly!"
	fi

	lustre_fail mds 0 0

	# no watchdog is triggered
	do_facet ost1 dmesg > $TMP/lustre-log-${TESTNAME}.log
	watchdog=$(awk '/Service thread pid/ && /was inactive/ \
			{ print; }' $TMP/lustre-log-${TESTNAME}.log)
	[ -z "$watchdog" ] || error "$watchdog"

	rm -f $TMP/lustre-log-${TESTNAME}.log

	# write should continue then fail with EDQUOT
	local count=0
	local c_size
	while [ true ]; do
		if ! ps -p ${DDPID} > /dev/null 2>&1; then break; fi
		if [ $count -ge 240 ]; then
			quota_error u $TSTUSR "dd not finished in $count secs"
		fi
		count=$((count + 1))
		if [ $((count % 30)) -eq 0 ]; then
			c_size=$(stat -c %s $TESTFILE)
			echo "Waiting $count secs. $c_size"
			$SHOW_QUOTA_USER
		fi
		sleep 1
	done

}
run_test 6 "Test dropping acquire request on master"

# quota reintegration (global index)
test_7a() {
	local TESTFILE=$DIR/$tdir/$tfile
	local LIMIT=20 # 20M

	[ "$SLOW" = "no" ] && LIMIT=5

	setup_quota_test || error "setup quota failed with $?"

	# make sure the system is clean
	local USED=$(getquota -u $TSTUSR global curspace)
	[ $USED -ne 0 ] && error "Used space($USED) for user $TSTUSR isn't 0."

	# make sure no granted quota on ost1
	set_ost_qtype $QTYPE || error "enable ost quota failed"
	resetquota -u $TSTUSR
	set_ost_qtype "none" || error "disable ost quota failed"

	local OSTUUID=$(ostuuid_from_index 0)
	USED=$(getquota -u $TSTUSR $OSTUUID bhardlimit)
	[ $USED -ne 0 ] &&
		error "limit($USED) on $OSTUUID for user $TSTUSR isn't 0"

	# create test file
	$SETSTRIPE $TESTFILE -c 1 -i 0 || error "setstripe $TESTFILE failed"
	chown $TSTUSR.$TSTUSR $TESTFILE || error "chown $TESTFILE failed"

	echo "Stop ost1..."
	stop ost1

	echo "Enable quota & set quota limit for $TSTUSR"
	set_ost_qtype $QTYPE || error "enable ost quota failed"
	$LFS setquota -u $TSTUSR -b 0 -B ${LIMIT}M -i 0 -I 0 $DIR ||
		error "set quota failed"

	echo "Start ost1..."
	start ost1 $(ostdevname 1) $OST_MOUNT_OPTS || error "start ost1 failed"
	quota_init

	wait_ost_reint $QTYPE || error "reintegration failed"

	# hardlimit should have been fetched by slave during global
	# reintegration, write will exceed quota
	$RUNAS $DD of=$TESTFILE count=$((LIMIT + 1)) oflag=sync &&
		quota_error u $TSTUSR "write success, but expect EDQUOT"

	rm -f $TESTFILE
	wait_delete_completed
	sync_all_data || true
	sleep 3

	echo "Stop ost1..."
	stop ost1

	$LFS setquota -u $TSTUSR -b 0 -B 0 -i 0 -I 0 $DIR ||
		error "clear quota failed"

	echo "Start ost1..."
	start ost1 $(ostdevname 1) $OST_MOUNT_OPTS || error "start ost1 failed"
	quota_init

	wait_ost_reint $QTYPE || error "reintegration failed"

	# hardlimit should be cleared on slave during reintegration
	$RUNAS $DD of=$TESTFILE count=$((LIMIT + 1)) oflag=sync ||
		quota_error u $TSTUSR "write error, but expect success"

}
run_test 7a "Quota reintegration (global index)"

# quota reintegration (slave index)
test_7b() {
	local LIMIT="100G"
	local TESTFILE=$DIR/$tdir/$tfile

	setup_quota_test || error "setup quota failed with $?"

	# make sure the system is clean
	local USED=$(getquota -u $TSTUSR global curspace)
	[ $USED -ne 0 ] && error "Used space($USED) for user $TSTUSR isn't 0."

	# make sure no granted quota on ost1
	set_ost_qtype $QTYPE || error "enable ost quota failed"
	resetquota -u $TSTUSR
	set_ost_qtype "none" || error "disable ost quota failed"

	local OSTUUID=$(ostuuid_from_index 0)
	USED=$(getquota -u $TSTUSR $OSTUUID bhardlimit)
	[ $USED -ne 0 ] &&
		error "limit($USED) on $OSTUUID for user $TSTUSR isn't 0"

	# create test file
	$SETSTRIPE $TESTFILE -c 1 -i 0 || error "setstripe $TESTFILE failed"
	chown $TSTUSR.$TSTUSR $TESTFILE || error "chown $TESTFILE failed"

	# consume some space to make sure the granted space will not
	# be released during reconciliation
	$RUNAS $DD of=$TESTFILE count=1 oflag=sync ||
		error "consume space failure, expect success"

	# define OBD_FAIL_QUOTA_EDQUOT 0xa02
	lustre_fail mds 0xa02

	set_ost_qtype $QTYPE || error "enable ost quota failed"
	$LFS setquota -u $TSTUSR -b 0 -B $LIMIT -i 0 -I 0 $DIR ||
		error "set quota failed"

	# ignore the write error
	$RUNAS $DD of=$TESTFILE count=1 seek=1 oflag=sync conv=notrunc

	local old_used=$(getquota -u $TSTUSR $OSTUUID bhardlimit)

	lustre_fail mds 0

	echo "Restart ost to trigger reintegration..."
	stop ost1
	start ost1 $(ostdevname 1) $OST_MOUNT_OPTS || error "start ost1 failed"
	quota_init

	wait_ost_reint $QTYPE || error "reintegration failed"

	USED=$(getquota -u $TSTUSR $OSTUUID bhardlimit)
	[ $USED -gt $old_used ] || error "limit on $OSTUUID $USED <= $old_used"

	cleanup_quota_test
	resetquota -u $TSTUSR
	$SHOW_QUOTA_USER
}
run_test 7b "Quota reintegration (slave index)"

# quota reintegration (restart mds during reintegration)
test_7c() {
	local LIMIT=20 # 20M
	local TESTFILE=$DIR/$tdir/$tfile

	[ "$SLOW" = "no" ] && LIMIT=5

	setup_quota_test || error "setup quota failed with $?"

	# make sure the system is clean
	local USED=$(getquota -u $TSTUSR global curspace)
	[ $USED -ne 0 ] && error "Used space($USED) for user $TSTUSR isn't 0."

	set_ost_qtype "none" || error "disable ost quota failed"
	$LFS setquota -u $TSTUSR -b 0 -B ${LIMIT}M -i 0 -I 0 $DIR ||
		error "set quota failed"

	# define OBD_FAIL_QUOTA_DELAY_REINT 0xa03
	lustre_fail ost 0xa03

	# enable ost quota
	set_ost_qtype $QTYPE || error "enable ost quota failed"
	# trigger reintegration
	local procf="osd-$(facet_fstype ost1).$FSNAME-OST*."
	procf=${procf}quota_slave.force_reint
	do_facet ost1 $LCTL set_param $procf=1 ||
		error "force reintegration failed"

	echo "Stop mds..."
	stop mds1

	lustre_fail ost 0

	echo "Start mds..."
	start mds1 $(mdsdevname 1) $MDS_MOUNT_OPTS
	quota_init

	# wait longer than usual to make sure the reintegration
	# is triggered by quota wb thread.
	wait_ost_reint $QTYPE 200 || error "reintegration failed"

	# hardlimit should have been fetched by slave during global
	# reintegration, write will exceed quota
	$RUNAS $DD of=$TESTFILE count=$((LIMIT + 1)) oflag=sync &&
		quota_error u $TSTUSR "write success, but expect EDQUOT"

	return 0
}
run_test 7c "Quota reintegration (restart mds during reintegration)"

# Quota reintegration (Transfer index in multiple bulks)
test_7d(){
	local TESTFILE=$DIR/$tdir/$tfile
	local TESTFILE1="$DIR/$tdir/$tfile"-1
	local limit=20 #20M

	setup_quota_test || error "setup quota failed with $?"

	set_ost_qtype "none" || error "disable ost quota failed"
	$LFS setquota -u $TSTUSR -B ${limit}M $DIR ||
		error "set quota for $TSTUSR failed"
	$LFS setquota -u $TSTUSR2 -B ${limit}M $DIR ||
		error "set quota for $TSTUSR2 failed"

	#define OBD_FAIL_OBD_IDX_READ_BREAK 0x608
	lustre_fail mds 0x608 0

	# enable quota to tirgger reintegration
	set_ost_qtype "u" || error "enable ost quota failed"
	wait_ost_reint "u" || error "reintegration failed"

	lustre_fail mds 0

	# hardlimit should have been fetched by slave during global
	# reintegration, write will exceed quota
	$RUNAS $DD of=$TESTFILE count=$((limit + 1)) oflag=sync &&
		quota_error u $TSTUSR "$TSTUSR write success, expect EDQUOT"

	$RUNAS2 $DD of=$TESTFILE1 count=$((limit + 1)) oflag=sync &&
		quota_error u $TSTUSR2 "$TSTUSR2 write success, expect EDQUOT"

	return 0
}
run_test 7d "Quota reintegration (Transfer index in multiple bulks)"

# quota reintegration (inode limits)
test_7e() {
	[ "$MDSCOUNT" -lt "2" ] && skip "needs >= 2 MDTs"

	# LU-2435: skip this quota test if underlying zfs version has not
	# supported native dnode accounting
	[ "$(facet_fstype mds1)" == "zfs" ] && {
		local F="feature@userobj_accounting"
		local pool=$(zpool_name mds1)
		local feature=$(do_facet mds1 $ZPOOL get -H $F $pool)

		[[ "$feature" != *" active "* ]] &&
			skip "requires zpool with active userobj_accounting"
	}

	local ilimit=$((1024 * 2)) # 2k inodes
	local TESTFILE=$DIR/${tdir}-1/$tfile

	setup_quota_test || error "setup quota failed with $?"

	# make sure the system is clean
	local USED=$(getquota -u $TSTUSR global curinodes)
	[ $USED -ne 0 ] && error "Used inode($USED) for user $TSTUSR isn't 0."

	# make sure no granted quota on mdt1
	set_mdt_qtype $QTYPE || error "enable mdt quota failed"
	resetquota -u $TSTUSR
	set_mdt_qtype "none" || error "disable mdt quota failed"

	local MDTUUID=$(mdtuuid_from_index $((MDSCOUNT - 1)))
	USED=$(getquota -u $TSTUSR $MDTUUID ihardlimit)
	[ $USED -ne 0 ] && error "limit($USED) on $MDTUUID for user" \
		"$TSTUSR isn't 0."

	echo "Stop mds${MDSCOUNT}..."
	stop mds${MDSCOUNT}

	echo "Enable quota & set quota limit for $TSTUSR"
	set_mdt_qtype $QTYPE || error "enable mdt quota failed"
	$LFS setquota -u $TSTUSR -b 0 -B 0 -i 0 -I $ilimit $DIR ||
		error "set quota failed"

	echo "Start mds${MDSCOUNT}..."
	start mds${MDSCOUNT} $(mdsdevname $MDSCOUNT) $MDS_MOUNT_OPTS
	quota_init

	wait_mdt_reint $QTYPE || error "reintegration failed"

	echo "create remote dir"
	$LFS mkdir -i $((MDSCOUNT - 1)) $DIR/${tdir}-1 ||
		error "create remote dir failed"
	chmod 0777 $DIR/${tdir}-1

	# hardlimit should have been fetched by slave during global
	# reintegration, create will exceed quota
	$RUNAS createmany -m $TESTFILE $((ilimit + 1)) &&
		quota_error u $TSTUSR "create succeeded, expect EDQUOT"

	$RUNAS unlinkmany $TESTFILE $ilimit || error "unlink files failed"
	wait_delete_completed
	sync_all_data || true

	echo "Stop mds${MDSCOUNT}..."
	stop mds${MDSCOUNT}

	$LFS setquota -u $TSTUSR -b 0 -B 0 -i 0 -I 0 $DIR ||
		error "clear quota failed"

	echo "Start mds${MDSCOUNT}..."
	start mds${MDSCOUNT} $(mdsdevname $MDSCOUNT) $MDS_MOUNT_OPTS
	quota_init

	wait_mdt_reint $QTYPE || error "reintegration failed"

	# hardlimit should be cleared on slave during reintegration
	$RUNAS createmany -m $TESTFILE $((ilimit + 1)) ||
		quota_error u $TSTUSR "create failed, expect success"

	$RUNAS unlinkmany $TESTFILE $((ilimit + 1)) || error "unlink failed"
	rmdir $DIR/${tdir}-1 || error "unlink remote dir failed"

}
run_test 7e "Quota reintegration (inode limits)"

# run dbench with quota enabled
test_8() {
	local BLK_LIMIT="100g" #100G
	local FILE_LIMIT=1000000

	setup_quota_test || error "setup quota failed with $?"

	set_mdt_qtype $QTYPE || error "enable mdt quota failed"
	set_ost_qtype $QTYPE || error "enable ost quota failed"

	echo "Set enough high limit for user: $TSTUSR"
	$LFS setquota -u $TSTUSR -b 0 -B $BLK_LIMIT -i 0 -I $FILE_LIMIT $DIR ||
		error "set user quota failed"
	echo "Set enough high limit for group: $TSTUSR"
	$LFS setquota -g $TSTUSR -b 0 -B $BLK_LIMIT -i 0 -I $FILE_LIMIT $DIR ||
		error "set group quota failed"
	if is_project_quota_supported; then
		change_project -sp $TSTPRJID $DIR/$tdir
		echo "Set enough high limit for project: $TSTPRJID"
		$LFS setquota -p $TSTPRJID -b 0 \
			-B $BLK_LIMIT -i 0 -I $FILE_LIMIT $DIR ||
			error "set project quota failed"
	fi

	local duration=""
	[ "$SLOW" = "no" ] && duration=" -t 120"
	$RUNAS bash rundbench -D $DIR/$tdir 3 $duration ||
		quota_error a $TSTUSR "dbench failed!"

	is_project_quota_supported && change_project -C $DIR/$tdir
}
run_test 8 "Run dbench with quota enabled"

# this check is just for test_9
OST0_MIN=4900000 #4.67G

check_whether_skip () {
	local OST0_SIZE=$($LFS df $DIR | awk '/\[OST:0\]/ {print $4}')
	log "OST0_SIZE: $OST0_SIZE  required: $OST0_MIN"
	if [ $OST0_SIZE -lt $OST0_MIN ]; then
		echo "WARN: OST0 has less than $OST0_MIN free, skip this test."
		return 0
	else
		return 1
	fi
}

# run for fixing bug10707, it needs a big room. test for 64bit
test_9() {
	local filesize=$((1024 * 9 / 2)) # 4.5G

	check_whether_skip && return 0

	setup_quota_test || error "setup quota failed with $?"

	set_ost_qtype "ug" || error "enable ost quota failed"

	local TESTFILE="$DIR/$tdir/$tfile-0"
	local BLK_LIMIT=100G #100G
	local FILE_LIMIT=1000000

	echo "Set block limit $BLK_LIMIT bytes to $TSTUSR.$TSTUSR"

	log "Set enough high limit(block:$BLK_LIMIT; file: $FILE_LIMIT)" \
		"for user: $TSTUSR"
	$LFS setquota -u $TSTUSR -b 0 -B $BLK_LIMIT -i 0 -I $FILE_LIMIT $DIR ||
		error "set user quota failed"

	log "Set enough high limit(block:$BLK_LIMIT; file: $FILE_LIMIT)" \
		"for group: $TSTUSR"
	$LFS setquota -g $TSTUSR -b 0 -B $BLK_LIMIT -i 0 -I $FILE_LIMIT $DIR ||
		error "set group quota failed"

	quota_show_check a u $TSTUSR
	quota_show_check a g $TSTUSR

	echo "Create test file"
	$SETSTRIPE $TESTFILE -c 1 -i 0 || error "setstripe $TESTFILE failed"
	chown $TSTUSR.$TSTUSR $TESTFILE || error "chown $TESTFILE failed"

	log "Write the big file of 4.5G ..."
	$RUNAS $DD of=$TESTFILE count=$filesize ||
		quota_error a $TSTUSR "write 4.5G file failure, expect success"

	$SHOW_QUOTA_USER
	$SHOW_QUOTA_GROUP

	cleanup_quota_test
	resetquota -u $TSTUSR
	resetquota -g $TSTUSR

	$SHOW_QUOTA_USER
	$SHOW_QUOTA_GROUP
}
run_test 9 "Block limit larger than 4GB (b10707)"

test_10() {
	local TESTFILE=$DIR/$tdir/$tfile

	setup_quota_test || error "setup quota failed with $?"

	# set limit to root user should fail
	$LFS setquota -u root -b 100G -B 500G -i 1K -I 1M $DIR &&
		error "set limit for root user successfully, expect failure"
	$LFS setquota -g root -b 1T -B 10T -i 5K -I 100M $DIR &&
		error "set limit for root group successfully, expect failure"
	$LFS setquota -p 0 -b 1T -B 10T -i 5K -I 100M $DIR &&
		error "set limit for project 0 successfully, expect failure"

	# root user can overrun quota
	set_ost_qtype "ug" || error "enable ost quota failed"

	$LFS setquota -u $TSTUSR -b 0 -B 2M -i 0 -I 0 $DIR ||
		error "set quota failed"
	quota_show_check b u $TSTUSR

	$SETSTRIPE $TESTFILE -c 1 || error "setstripe $TESTFILE failed"
	chown $TSTUSR.$TSTUSR $TESTFILE || error "chown $TESTFILE failed"

	runas -u 0 -g 0 $DD of=$TESTFILE count=3 oflag=sync ||
		error "write failure, expect success"

}
run_test 10 "Test quota for root user"

test_11() {
	local TESTFILE=$DIR/$tdir/$tfile
	setup_quota_test || error "setup quota failed with $?"

	set_mdt_qtype "ug" || error "enable mdt quota failed"
	$LFS setquota -u $TSTUSR -b 0 -B 0 -i 0 -I 1 $DIR ||
		error "set quota failed"

	touch "$TESTFILE"-0 || error "touch $TESTFILE-0 failed"
	touch "$TESTFILE"-1 || error "touch $TESTFILE-0 failed"

	chown $TSTUSR.$TSTUSR "$TESTFILE"-0 || error "chown $TESTFILE-0 failed"
	chown $TSTUSR.$TSTUSR "$TESTFILE"-1 || error "chown $TESTFILE-1 failed"

	$SHOW_QUOTA_USER
	local USED=$(getquota -u $TSTUSR global curinodes)
	[ $USED -ge 2 ] || error "Used inodes($USED) is less than 2"

}
run_test 11 "Chown/chgrp ignores quota"

test_12a() {
	[ "$OSTCOUNT" -lt "2" ] && skip "needs >= 2 OSTs"

	local blimit=22 # 22M
	local blk_cnt=$((blimit - 5))
	local TESTFILE0="$DIR/$tdir/$tfile"-0
	local TESTFILE1="$DIR/$tdir/$tfile"-1

	setup_quota_test || error "setup quota failed with $?"

	set_ost_qtype "u" || error "enable ost quota failed"
	quota_show_check b u $TSTUSR

	$LFS setquota -u $TSTUSR -b 0 -B "$blimit"M -i 0 -I 0 $DIR ||
		error "set quota failed"

	$SETSTRIPE $TESTFILE0 -c 1 -i 0 || error "setstripe $TESTFILE0 failed"
	$SETSTRIPE $TESTFILE1 -c 1 -i 1 || error "setstripe $TESTFILE1 failed"
	chown $TSTUSR.$TSTUSR $TESTFILE0 || error "chown $TESTFILE0 failed"
	chown $TSTUSR.$TSTUSR $TESTFILE1 || error "chown $TESTFILE1 failed"

	echo "Write to ost0..."
	$RUNAS $DD of=$TESTFILE0 count=$blk_cnt oflag=sync ||
		quota_error a $TSTUSR "dd failed"

	echo "Write to ost1..."
	$RUNAS $DD of=$TESTFILE1 count=$blk_cnt oflag=sync &&
		quota_error a $TSTUSR "dd succeed, expect EDQUOT"

	echo "Free space from ost0..."
	rm -f $TESTFILE0
	wait_delete_completed
	sync_all_data || true

	echo "Write to ost1 after space freed from ost0..."
	$RUNAS $DD of=$TESTFILE1 count=$blk_cnt oflag=sync ||
		quota_error a $TSTUSR "rebalancing failed"

}
run_test 12a "Block quota rebalancing"

test_12b() {
	[ "$MDSCOUNT" -lt "2" ] && skip "needs >= 2 MDTs"

	local ilimit=$((1024 * 2)) # 2k inodes
	local TESTFILE0=$DIR/$tdir/$tfile
	local TESTFILE1=$DIR/${tdir}-1/$tfile

	setup_quota_test || error "setup quota failed with $?"

	$LFS mkdir -i 1 $DIR/${tdir}-1 || error "create remote dir failed"
	chmod 0777 $DIR/${tdir}-1

	set_mdt_qtype "u" || error "enable mdt quota failed"
	quota_show_check f u $TSTUSR

	$LFS setquota -u $TSTUSR -b 0 -B 0 -i 0 -I $ilimit $DIR ||
		error "set quota failed"

	echo "Create $ilimit files on mdt0..."
	$RUNAS createmany -m $TESTFILE0 $ilimit ||
		quota_error u $TSTUSR "create failed, but expect success"

	echo "Create files on mdt1..."
	$RUNAS createmany -m $TESTFILE1 1 &&
		quota_error a $TSTUSR "create succeeded, expect EDQUOT"

	echo "Free space from mdt0..."
	$RUNAS unlinkmany $TESTFILE0 $ilimit || error "unlink mdt0 files failed"
	wait_delete_completed
	sync_all_data || true

	echo "Create files on mdt1 after space freed from mdt0..."
	$RUNAS createmany -m $TESTFILE1 $((ilimit / 2)) ||
		quota_error a $TSTUSR "rebalancing failed"

	$RUNAS unlinkmany $TESTFILE1 $((ilimit / 2)) ||
		error "unlink mdt1 files failed"
	rmdir $DIR/${tdir}-1 || error "unlink remote dir failed"

}
run_test 12b "Inode quota rebalancing"

test_13(){
	local TESTFILE=$DIR/$tdir/$tfile
	# the name of lwp on ost1 name is MDT0000-lwp-OST0000
	local procf="ldlm.namespaces.*MDT0000-lwp-OST0000.lru_size"

	setup_quota_test || error "setup quota failed with $?"

	set_ost_qtype "u" || error "enable ost quota failed"
	quota_show_check b u $TSTUSR

	$LFS setquota -u $TSTUSR -b 0 -B 10M -i 0 -I 0 $DIR ||
		error "set quota failed"
	$SETSTRIPE $TESTFILE -c 1 -i 0 || error "setstripe $TESTFILE failed"
	chown $TSTUSR.$TSTUSR $TESTFILE || error "chown $TESTFILE failed"

	# clear the locks in cache first
	do_facet ost1 $LCTL set_param -n $procf=clear
	local nlock=$(do_facet ost1 $LCTL get_param -n $procf)
	[ $nlock -eq 0 ] || error "$nlock cached locks"

	# write to acquire the per-ID lock
	$RUNAS $DD of=$TESTFILE count=1 oflag=sync ||
		quota_error a $TSTUSR "dd failed"

	nlock=$(do_facet ost1 $LCTL get_param -n $procf)
	[ $nlock -eq 1 ] || error "lock count($nlock) isn't 1"

	# clear quota doesn't trigger per-ID lock cancellation
	resetquota -u $TSTUSR
	nlock=$(do_facet ost1 $LCTL get_param -n $procf)
	[ $nlock -eq 1 ] || error "per-ID lock is lost on quota clear"

	# clear the per-ID lock
	do_facet ost1 $LCTL set_param -n $procf=clear
	nlock=$(do_facet ost1 $LCTL get_param -n $procf)
	[ $nlock -eq 0 ] || error "per-ID lock isn't cleared"

	# spare quota should be released
	local OSTUUID=$(ostuuid_from_index 0)
	local limit=$(getquota -u $TSTUSR $OSTUUID bhardlimit)
	local space=$(getquota -u $TSTUSR $OSTUUID curspace)
	[ $limit -le $space ] ||
		error "spare quota isn't released, limit:$limit, space:$space"
}
run_test 13 "Cancel per-ID lock in the LRU list"

test_15(){
	local LIMIT=$((24 * 1024 * 1024 * 1024 * 1024)) # 24 TB

	wait_delete_completed
	sync_all_data || true

	# test for user
	$LFS setquota -u $TSTUSR -b 0 -B $LIMIT -i 0 -I 0 $DIR ||
		error "set user quota failed"
	local TOTAL_LIMIT=$(getquota -u $TSTUSR global bhardlimit)
	[ $TOTAL_LIMIT -eq $LIMIT ] ||
		error "(user) limit:$TOTAL_LIMIT, expect:$LIMIT, failed!"
	resetquota -u $TSTUSR

	# test for group
	$LFS setquota -g $TSTUSR -b 0 -B $LIMIT -i 0 -I 0 $DIR ||
		error "set group quota failed"
	TOTAL_LIMIT=$(getquota -g $TSTUSR global bhardlimit)
	[ $TOTAL_LIMIT -eq $LIMIT ] ||
		error "(group) limits:$TOTAL_LIMIT, expect:$LIMIT, failed!"
	resetquota -g $TSTUSR
}
run_test 15 "Set over 4T block quota"

test_16a()
{
	(( $CLIENT_VERSION < $(version_code 2.12.8.18) )) &&
		skip "Not supported Lustre client before 2.12.8.18"

	setup_quota_test || error "setup quota failed with $?"

	# make sure the system is clean
	local used=$(getquota -u $TSTUSR global curspace)
	[[ "$used" == "0" ]] || error "Used space($used) for user $TSTUSR not 0"

	$LFS setquota -u $TSTUSR -B 500M -I 10K $MOUNT ||
		error "failed to set quota for user $TSTUSR"
	$LFS setquota -g $TSTUSR -B 500M -I 10K $MOUNT ||
		error "failed to set quota for group $TSTUSR"

	stack_trap "rm -f $DIR/$tdir/$tfile"
	$RUNAS $DD of=$DIR/$tdir/$tfile bs=1M count=50 ||
		quota_error u $TSTUSR "write failure"

	$LFS quota -u $TSTUSR $MOUNT ||
		quota_error u $TSTUSR "failed to get quota"

	local OSC=$($LCTL dl | grep OST0000-osc-[^M] | awk '{print $4}')

	$LCTL --device %$OSC deactivate
	stack_trap "$LCTL --device %$OSC activate"

	$LFS quota -v -u $TSTUSR $MOUNT ||
		quota_error u $TSTUSR "failed to get quota after deactivate OSC"
	$LFS quota -v -g $TSTUSR $MOUNT ||
		quota_error g $TSTUSR "failed to get quota after deactivate OSC"

	(( $MDSCOUNT > 1 )) || return 0

	local MDC=$($LCTL dl | grep MDT0001-mdc-[^M] | awk '{print $4}')

	$LCTL --device %$MDC deactivate
	stack_trap "$LCTL --device %$MDC activate"

	$LFS quota -v -u $TSTUSR $MOUNT ||
		quota_error u $TSTUSR "failed to get quota after deactivate MDC"
	$LFS quota -v -g $TSTUSR $MOUNT ||
		quota_error g $TSTUSR "failed to get quota after deactivate OSC"
}
run_test 16a "lfs quota should skip the inactive MDT/OST"

cleanup_16b()
{
	stopall
	formatall
	setupall
}

test_16b()
{
	(( $CLIENT_VERSION < $(version_code 2.12.8.18) )) &&
		skip "Not supported Lustre client before 2.12.8.18"

	(( $MDSCOUNT >= 3 )) || skip "needs >= 3 MDTs"

	stopall
	if ! combined_mgs_mds ; then
		format_mgs
		start_mgs
	fi

	add mds1 $(mkfs_opts mds1 $(mdsdevname 1)) --index=0 --reformat \
		$(mdsdevname 1) $(mdsvdevname 1)
	add mds2 $(mkfs_opts mds2 $(mdsdevname 2)) --index=1 --reformat \
		$(mdsdevname 2) $(mdsvdevname 2)
	add mds3 $(mkfs_opts mds3 $(mdsdevname 3)) --index=100 --reformat \
		$(mdsdevname 3) $(mdsvdevname 3)

	add ost1 $(mkfs_opts ost1 $(ostdevname 1)) --index=0 --reformat \
		$(ostdevname 1) $(ostvdevname 1)
	add ost2 $(mkfs_opts ost2 $(ostdevname 2)) --index=100 --reformat \
		$(ostdevname 2) $(ostvdevname 2)

	stack_trap cleanup_16b

	start mds1 $(mdsdevname 1) $MDS_MOUNT_OPTS || error "MDT1 start failed"
	start mds2 $(mdsdevname 2) $MDS_MOUNT_OPTS || error "MDT2 start failed"
	start mds3 $(mdsdevname 3) $MDS_MOUNT_OPTS || error "MDT3 start failed"
	start ost1 $(ostdevname 1) $OST_MOUNT_OPTS || error "OST1 start failed"
	start ost2 $(ostdevname 2) $OST_MOUNT_OPTS || error "OST2 start failed"

	mount_client $MOUNT || error "Unable to mount client"

	setup_quota_test || error "setup quota failed with $?"

	$LFS setquota -u $TSTUSR -B 100M -I 10K $MOUNT ||
		error "failed to set quota for user $TSTUSR"
	$LFS setquota -g $TSTUSR -B 100M -I 10K $MOUNT ||
		error "failed to set quota for group $TSTUSR"

	$RUNAS $DD of=$DIR/$tdir/$tfile bs=1M count=10 ||
		quota_error u $TSTUSR "write failure"

	cnt=$($LFS quota -v -u $TSTUSR $MOUNT | grep -ce "^$FSNAME-[MD|OS]T*")
	[ $cnt -le 5 ] || quota_error u $TSTUSR "failed to get user quota"
	cnt=$($LFS quota -v -g $TSTUSR $MOUNT | grep -ce "^$FSNAME-[MD|OS]T*")
	[ $cnt -le 5 ] || quota_error g $TSTUSR "failed to get group quota"
}
run_test 16b "lfs quota should skip the nonexistent MDT/OST"

test_17sub() {
	local err_code=$1
	local BLKS=1    # 1M less than limit
	local TESTFILE=$DIR/$tdir/$tfile

	setup_quota_test || error "setup quota failed with $?"

	# make sure the system is clean
	local USED=$(getquota -u $TSTUSR global curspace)
	[ $USED -ne 0 ] && error "Used space($USED) for user $TSTUSR isn't 0."

	set_ost_qtype "ug" || error "enable ost quota failed"
	# make sure no granted quota on ost
	resetquota -u $TSTUSR
	$LFS setquota -u $TSTUSR -b 0 -B 10M -i 0 -I 0 $DIR ||
		error "set quota failed"

	quota_show_check b u $TSTUSR

	#define OBD_FAIL_QUOTA_RECOVERABLE_ERR 0xa04
	lustre_fail mds 0xa04 $err_code

	# write in background
	$RUNAS $DD of=$TESTFILE count=$BLKS oflag=direct &
	local DDPID=$!

	sleep 2
	# write should be blocked and never finished
	if ! ps -p $DDPID  > /dev/null 2>&1; then
		lustre_fail mds 0 0
		quota_error u $TSTUSR "write finished incorrectly!"
	fi

	lustre_fail mds 0 0

	local count=0
	local timeout=30
	while [ true ]; do
		if ! ps -p ${DDPID} > /dev/null 2>&1; then break; fi
		count=$((count+1))
		if [ $count -gt $timeout ]; then
			quota_error u $TSTUSR "dd is not finished!"
		fi
		sleep 1
	done

	sync; sync_all_data || true

	USED=$(getquota -u $TSTUSR global curspace)
	[ $USED -ge $((BLKS * 1024)) ] || quota_error u $TSTUSR \
		"Used space(${USED}K) is less than ${BLKS}M"

	cleanup_quota_test
}

# DQACQ return recoverable error
test_17() {
	echo "DQACQ return -ENOLCK"
	#define ENOLCK  37
	test_17sub 37 || error "Handle -ENOLCK failed"

	echo "DQACQ return -EAGAIN"
	#define EAGAIN  11
	test_17sub 11 || error "Handle -EAGAIN failed"

	echo "DQACQ return -ETIMEDOUT"
	#define ETIMEDOUT 110
	test_17sub 110 || error "Handle -ETIMEDOUT failed"

	echo "DQACQ return -ENOTCONN"
	#define ENOTCONN 107
	test_17sub 107 || error "Handle -ENOTCONN failed"
}

run_test 17 "DQACQ return recoverable error"

test_18_sub () {
	local io_type=$1
	local blimit="200m" # 200M
	local TESTFILE="$DIR/$tdir/$tfile"

	setup_quota_test || error "setup quota failed with $?"

	set_ost_qtype "u" || error "enable ost quota failed"
	log "User quota (limit: $blimit)"
	$LFS setquota -u $TSTUSR -b 0 -B $blimit -i 0 -I 0 $MOUNT ||
		error "set quota failed"
	quota_show_check b u $TSTUSR

	$SETSTRIPE $TESTFILE -i 0 -c 1 || error "setstripe $TESTFILE failed"
	chown $TSTUSR.$TSTUSR $TESTFILE || error "chown $TESTFILE failed"

	local timeout=$(sysctl -n lustre.timeout)

	if [ $io_type = "directio" ]; then
		log "Write 100M (directio) ..."
		$RUNAS $DD of=$TESTFILE count=100 oflag=direct &
	else
		log "Write 100M (buffered) ..."
		$RUNAS $DD of=$TESTFILE count=100 &
	fi
	local DDPID=$!

	replay_barrier $SINGLEMDS
	log "Fail mds for $((2 * timeout)) seconds"
	fail $SINGLEMDS $((2 * timeout))

	local count=0
	if at_is_enabled; then
		timeout=$(at_max_get mds)
	else
		timeout=$(lctl get_param -n timeout)
	fi

	while [ true ]; do
		if ! ps -p ${DDPID} > /dev/null 2>&1; then break; fi
		if [ $((++count % (2 * timeout) )) -eq 0 ]; then
			log "it took $count second"
		fi
		sleep 1
	done

	log "(dd_pid=$DDPID, time=$count, timeout=$timeout)"
	sync
	cancel_lru_locks mdc
	cancel_lru_locks osc
	$SHOW_QUOTA_USER

	local testfile_size=$(stat -c %s $TESTFILE)
	if [ $testfile_size -ne $((BLK_SZ * 1024 * 100)) ] ; then
		quota_error u $TSTUSR "expect $((BLK_SZ * 1024 * 100))," \
			"got ${testfile_size}. Verifying file failed!"
	fi
	cleanup_quota_test
}

# test when mds does failover, the ost still could work well
# this test shouldn't trigger watchdog b=14840
test_18() {
	# Clear dmesg so watchdog is not triggered by previous
	# test output
	do_facet ost1 dmesg -c > /dev/null

	test_18_sub normal
	test_18_sub directio

	# check if watchdog is triggered
	do_facet ost1 dmesg > $TMP/lustre-log-${TESTNAME}.log
	local watchdog=$(awk '/Service thread pid/ && /was inactive/ \
			{ print; }' $TMP/lustre-log-${TESTNAME}.log)
	[ -z "$watchdog" ] || error "$watchdog"
	rm -f $TMP/lustre-log-${TESTNAME}.log
}
run_test 18 "MDS failover while writing, no watchdog triggered (b14840)"

test_19() {
	local blimit=5 # 5M
	local TESTFILE=$DIR/$tdir/$tfile

	setup_quota_test || error "setup quota failed with $?"

	set_ost_qtype $QTYPE || error "enable ost quota failed"

	# bind file to a single OST
	$SETSTRIPE -c 1 $TESTFILE || error "setstripe $TESTFILE failed"
	chown $TSTUSR.$TSTUSR $TESTFILE || error "chown $TESTFILE failed"

	echo "Set user quota (limit: ${blimit}M)"
	$LFS setquota -u $TSTUSR -b 0 -B "$blimit"M -i 0 -I 0 $MOUNT ||
		error "set user quota failed"
	quota_show_check b u $TSTUSR
	echo "Update quota limits"
	$LFS setquota -u $TSTUSR -b 0 -B "$blimit"M -i 0 -I 0 $MOUNT ||
		error "set group quota failed"
	quota_show_check b u $TSTUSR

	# first wirte might be cached
	$RUNAS $DD of=$TESTFILE count=$((blimit + 1))
	cancel_lru_locks osc
	$SHOW_QUOTA_USER
	$RUNAS $DD of=$TESTFILE count=$((blimit + 1)) seek=$((blimit + 1)) &&
		quota_error u $TSTUSR "Write success, expect failure"
	$SHOW_QUOTA_USER

}
run_test 19 "Updating admin limits doesn't zero operational limits(b14790)"

test_20() { # b15754
	local LSTR=(2g 1t 4k 3m) # limits strings
	# limits values
	local LVAL=($((2*1024*1024)) $((1*1024*1024*1024)) $((4*1024)) \
		    $((3*1024*1024)))

	resetquota -u $TSTUSR

	$LFS setquota -u $TSTUSR --block-softlimit ${LSTR[0]} \
		$MOUNT || error "could not set quota limits"
	$LFS setquota -u $TSTUSR --block-hardlimit ${LSTR[1]} \
				--inode-softlimit ${LSTR[2]} \
				--inode-hardlimit ${LSTR[3]} \
				$MOUNT || error "could not set quota limits"

	[ "$(getquota -u $TSTUSR global bsoftlimit)" = "${LVAL[0]}" ] ||
		error "bsoftlimit was not set properly"
	[ "$(getquota -u $TSTUSR global bhardlimit)" = "${LVAL[1]}" ] ||
		error "bhardlimit was not set properly"
	[ "$(getquota -u $TSTUSR global isoftlimit)" = "${LVAL[2]}" ] ||
		error "isoftlimit was not set properly"
	[ "$(getquota -u $TSTUSR global ihardlimit)" = "${LVAL[3]}" ] ||
		error "ihardlimit was not set properly"

	resetquota -u $TSTUSR
}
run_test 20 "Test if setquota specifiers work properly (b15754)"

test_21_sub() {
	local testfile=$1
	local blk_number=$2
	local seconds=$3

	local time=$(($(date +%s) + seconds))
	while [ $(date +%s) -lt $time ]; do
		$RUNAS $DD of=$testfile count=$blk_number > /dev/null 2>&1
	done
}

# run for fixing bug16053, setquota shouldn't fail when writing and
# deleting are happening
test_21() {
	local TESTFILE="$DIR/$tdir/$tfile"
	local BLIMIT=10 # 10G
	local ILIMIT=1000000

	setup_quota_test || error "setup quota failed with $?"

	set_ost_qtype $QTYPE || error "Enable ost quota failed"

	log "Set limit(block:${BLIMIT}G; file:$ILIMIT) for user: $TSTUSR"
	$LFS setquota -u $TSTUSR -b 0 -B ${BLIMIT}G -i 0 -I $ILIMIT $MOUNT ||
		error "set user quota failed"
	log "Set limit(block:${BLIMIT}G; file:$ILIMIT) for group: $TSTUSR"
	$LFS setquota -g $TSTUSR -b 0 -B $BLIMIT -i 0 -I $ILIMIT $MOUNT ||
		error "set group quota failed"
	if is_project_quota_supported; then
		log "Set limit(block:${BLIMIT}G; file:$LIMIT) for " \
			"project: $TSTPRJID"
		$LFS setquota -p $TSTPRJID -b 0 -B $BLIMIT -i 0 -I $ILIMIT \
			 $MOUNT || error "set project quota failed"
	fi

	# repeat writing on a 1M file
	test_21_sub ${TESTFILE}_1 1 30 &
	local DDPID1=$!
	# repeat writing on a 128M file
	test_21_sub ${TESTFILE}_2 128 30 &
	local DDPID2=$!

	local time=$(($(date +%s) + 30))
	local i=1
	while [ $(date +%s) -lt $time ]; do
		log "Set quota for $i times"
		$LFS setquota -u $TSTUSR -b 0 -B "$((BLIMIT + i))G" -i 0 \
			-I $((ILIMIT + i)) $MOUNT ||
				error "Set user quota failed"
		$LFS setquota -g $TSTUSR -b 0 -B "$((BLIMIT + i))G" -i 0 \
			-I $((ILIMIT + i)) $MOUNT ||
				error "Set group quota failed"
		if is_project_quota_supported; then
			$LFS setquota -p $TSTPRJID -b 0 -B \
			"$((BLIMIT + i))G"  -i 0 -I $((ILIMIT + i)) $MOUNT ||
				error "Set project quota failed"
		fi
		i=$((i+1))
		sleep 1
	done

	local count=0
	while [ true ]; do
		if ! ps -p ${DDPID1} > /dev/null 2>&1; then break; fi
		count=$((count+1))
		if [ $count -gt 60 ]; then
			quota_error a $TSTUSR "dd should be finished!"
		fi
		sleep 1
	done
	echo "(dd_pid=$DDPID1, time=$count)successful"

	count=0
	while [ true ]; do
		if ! ps -p ${DDPID2} > /dev/null 2>&1; then break; fi
		count=$((count+1))
		if [ $count -gt 60 ]; then
			quota_error a $TSTUSR "dd should be finished!"
		fi
		sleep 1
	done
	echo "(dd_pid=$DDPID2, time=$count)successful"

}
run_test 21 "Setquota while writing & deleting (b16053)"

# enable/disable quota enforcement permanently
test_22() {
	echo "Set both mdt & ost quota type as ug"
	local qtype="ug"
	is_project_quota_supported && qtype=$QTYPE
	set_mdt_qtype $qtype || error "enable mdt quota failed"
	set_ost_qtype $qtype || error "enable ost quota failed"

	echo "Restart..."
	stopall || error "failed to stopall (1)"
	mount
	setupall

	echo "Verify if quota is enabled"
	local qtype1=$(mdt_quota_type)
	[ $qtype1 != $qtype ] && error "mdt quota setting is lost"
	qtype=$(ost_quota_type)
	[ $qtype1 != $qtype ] && error "ost quota setting is lost"

	echo "Set both mdt & ost quota type as none"
	set_mdt_qtype "none" || error "disable mdt quota failed"
	set_ost_qtype "none" || error "disable ost quota failed"

	echo "Restart..."
	stopall || error "failed to stopall (2)"
	mount
	setupall
	quota_init

	echo "Verify if quota is disabled"
	qtype=$(mdt_quota_type)
	[ $qtype != "none" ] && error "mdt quota setting is lost"
	qtype=$(ost_quota_type)
	[ $qtype != "none" ] && error "ost quota setting is lost"

	return 0
}
run_test 22 "enable/disable quota by 'lctl conf_param/set_param -P'"

test_23_sub() {
	local TESTFILE="$DIR/$tdir/$tfile"
	local LIMIT=$1

	setup_quota_test || error "setup quota failed with $?"

	set_ost_qtype $QTYPE || error "Enable ost quota failed"

	# test for user
	log "User quota (limit: $LIMIT MB)"
	$LFS setquota -u $TSTUSR -b 0 -B "$LIMIT"M -i 0 -I 0 $DIR ||
		error "set quota failed"
	quota_show_check b u $TSTUSR

	$SETSTRIPE $TESTFILE -c 1 -i 0 || error "setstripe $TESTFILE failed"
	chown $TSTUSR.$TSTUSR $TESTFILE || error "chown $TESTFILE failed"

	log "Step1: trigger EDQUOT with O_DIRECT"
	log "Write half of file"
	$RUNAS $DD of=$TESTFILE count=$((LIMIT/2)) oflag=direct ||
		quota_error u $TSTUSR "(1) Write failure, expect success." \
			"limit=$LIMIT"
	log "Write out of block quota ..."
	$RUNAS $DD of=$TESTFILE count=$((LIMIT/2 + 1)) seek=$((LIMIT/2)) \
		oflag=direct conv=notrunc &&
		quota_error u $TSTUSR "(2) Write success, expect EDQUOT." \
			"limit=$LIMIT"
	log "Step1: done"

	log "Step2: rewrite should succeed"
	$RUNAS $DD of=$TESTFILE count=1 oflag=direct conv=notrunc||
		quota_error u $TSTUSR "(3) Write failure, expect success." \
			"limit=$LIMIT"
	log "Step2: done"

	cleanup_quota_test

	local OST0_UUID=$(ostuuid_from_index 0)
	local OST0_QUOTA_USED=$(getquota -u $TSTUSR $OST0_UUID curspace)
	[ $OST0_QUOTA_USED -ne 0 ] &&
		($SHOW_QUOTA_USER; \
		quota_error u $TSTUSR "quota isn't released")
	$SHOW_QUOTA_USER
}

test_23() {
	[ $(facet_fstype ost1) == "zfs" ] &&
		skip "Overwrite in place is not guaranteed to be " \
		"space neutral on ZFS"

	local OST0_MIN=$((6 * 1024)) # 6MB, extra space for meta blocks.
	check_whether_skip && return 0
	log "run for 4MB test file"
	test_23_sub 4

	OST0_MIN=$((60 * 1024)) # 60MB, extra space for meta blocks.
	check_whether_skip && return 0
	log "run for 40MB test file"
	test_23_sub 40
}
run_test 23 "Quota should be honored with directIO (b16125)"

test_24() {
	local blimit=5 # 5M
	local TESTFILE="$DIR/$tdir/$tfile"

	setup_quota_test || error "setup quota failed with $?"

	set_ost_qtype $QTYPE || error "enable ost quota failed"

	# bind file to a single OST
	$SETSTRIPE -c 1 $TESTFILE || error "setstripe $TESTFILE failed"
	chown $TSTUSR.$TSTUSR $TESTFILE || error "chown $TESTFILE failed"

	echo "Set user quota (limit: ${blimit}M)"
	$LFS setquota -u $TSTUSR -b 0 -B "$blimit"M -i 0 -I 0 $MOUNT ||
		error "set quota failed"

	# overrun quota by root user
	runas -u 0 -g 0 $DD of=$TESTFILE count=$((blimit + 1)) ||
		error "write failure, expect success"
	cancel_lru_locks osc
	sync_all_data || true

	$SHOW_QUOTA_USER | grep '*' || error "no matching *"

}
run_test 24 "lfs draws an asterix when limit is reached (b16646)"

test_27a() { # b19612
	$LFS quota $TSTUSR $DIR &&
		error "lfs succeeded with no type, but should have failed"
	$LFS setquota $TSTUSR $DIR &&
		error "lfs succeeded with no type, but should have failed"
	return 0
}
run_test 27a "lfs quota/setquota should handle wrong arguments (b19612)"

test_27b() { # b20200
	$LFS setquota -u $TSTID -b 1000 -B 1000 -i 1000 -I 1000 $DIR ||
		error "lfs setquota failed with uid argument"
	$LFS setquota -g $TSTID -b 1000 -B 1000 -i 1000 -I 1000 $DIR ||
		error "lfs stequota failed with gid argument"
	if is_project_quota_supported; then
		$LFS setquota -p $TSTPRJID -b 1000 -B 1000 -i 1000 -I \
			1000 $DIR || error \
				"lfs stequota failed with projid argument"
	fi
	$SHOW_QUOTA_USERID || error "lfs quota failed with uid argument"
	$SHOW_QUOTA_GROUPID || error "lfs quota failed with gid argument"
	if is_project_quota_supported; then
		$SHOW_QUOTA_PROJID ||
			error "lfs quota failed with projid argument"
	fi
	resetquota -u $TSTUSR
	resetquota -g $TSTUSR
	resetquota -p $TSTPRJID
	return 0
}
run_test 27b "lfs quota/setquota should handle user/group/project ID (b20200)"

test_27c() {
	local limit

	$LFS setquota -u $TSTID -b 30M -B 3T $DIR ||
		error "lfs setquota failed"

	limit=$($LFS quota -u $TSTID -v -h $DIR | grep $DIR | awk '{print $3}')
	[ $limit != "30M" ] && error "softlimit $limit isn't human-readable"
	limit=$($LFS quota -u $TSTID -v -h $DIR | grep $DIR | awk '{print $4}')
	[ $limit != "3T" ] && error "hardlimit $limit isn't human-readable"

	$LFS setquota -u $TSTID -b 1500M -B 18500G $DIR ||
		error "lfs setquota for $TSTID failed"

	limit=$($LFS quota -u $TSTID -v -h $DIR | grep $DIR | awk '{print $3}')
	[ $limit != "1.465G" ] && error "wrong softlimit $limit"
	limit=$($LFS quota -u $TSTID -v -h $DIR | grep $DIR | awk '{print $4}')
	[ $limit != "18.07T" ] && error "wrong hardlimit $limit"

	$LFS quota -u $TSTID -v -h $DIR | grep -q "Total allocated" ||
		error "total allocated inode/block limit not printed"

	resetquota -u $TSTUSR
}
run_test 27c "lfs quota should support human-readable output"

test_27d() {
	local softlimit=1.5
	local hardlimit=2.3
	local limit

	$LFS setquota -u $TSTID -b ${softlimit}p -B ${hardlimit}P $DIR ||
		error "set fraction block limit failed"
	limit=$($LFS quota -u $TSTID -h $DIR | grep $DIR | awk '{print $3}')
	[ $limit == ${softlimit}P ] || error "get fraction softlimit failed"
	limit=$($LFS quota -u $TSTID -h $DIR | grep $DIR | awk '{print $4}')
	[ $limit == ${hardlimit}P ] || error "get fraction hardlimit failed"

	resetquota -u $TSTUSR
}
run_test 27d "lfs setquota should support fraction block limit"

test_30() {
	local LIMIT=4 # 4MB
	local TESTFILE="$DIR/$tdir/$tfile"
	local GRACE=10

	setup_quota_test || error "setup quota failed with $?"

	set_ost_qtype "u" || error "enable ost quota failed"

	$SETSTRIPE $TESTFILE -i 0 -c 1 || error "setstripe $TESTFILE failed"
	chown $TSTUSR.$TSTUSR $TESTFILE || error "chown $TESTFILE failed"

	$LFS setquota -t -u --block-grace $GRACE --inode-grace \
		$MAX_IQ_TIME $DIR || error "set grace time failed"
	$LFS setquota -u $TSTUSR -b ${LIMIT}M -B 0 -i 0 -I 0 $DIR ||
		error "set quota failed"
	$RUNAS $DD of=$TESTFILE count=$((LIMIT * 2)) || true
	cancel_lru_locks osc
	sleep $GRACE
	$LFS setquota -u $TSTUSR -B 0 $DIR || error "clear quota failed"
	# over-quota flag has not yet settled since we do not trigger async
	# events based on grace time period expiration
	$SHOW_QUOTA_USER
	$RUNAS $DD of=$TESTFILE conv=notrunc oflag=append count=4 || true
	cancel_lru_locks osc
	# now over-quota flag should be settled and further writes should fail
	$SHOW_QUOTA_USER
	$RUNAS $DD of=$TESTFILE conv=notrunc oflag=append count=4 &&
		error "grace times were reset"

	resetquota -u $TSTUSR
	$LFS setquota -t -u --block-grace $MAX_DQ_TIME --inode-grace \
		$MAX_IQ_TIME $DIR || error "restore grace time failed"
}
run_test 30 "Hard limit updates should not reset grace times"

# basic usage tracking for user & group
test_33() {
	local INODES=10 # 10 files
	local BLK_CNT=2 # of 2M each
	local TOTAL_BLKS=$((INODES * BLK_CNT * 1024))

	setup_quota_test || error "setup quota failed with $?"

	# make sure the system is clean
	local USED=$(getquota -u $TSTID global curspace)
	[ $USED -ne 0 ] &&
		error "Used space ($USED) for user $TSTID isn't 0."
	USED=$(getquota -g $TSTID global curspace)
	[ $USED -ne 0 ] &&
		error "Used space ($USED) for group $TSTID isn't 0."
	if is_project_quota_supported; then
		USED=$(getquota -p $TSTPRJID global curspace)
		[ $USED -ne 0 ] && error \
			"Used space ($USED) for project $TSTPRJID isn't 0."
	fi

	echo "Write files..."
	for i in $(seq 0 $INODES); do
		$RUNAS $DD of=$DIR/$tdir/$tfile-$i count=$BLK_CNT 2>/dev/null ||
			error "write failed"
			is_project_quota_supported &&
				change_project -p $TSTPRJID $DIR/$tdir/$tfile-$i
		echo "Iteration $i/$INODES completed"
	done
	cancel_lru_locks osc

	echo "Wait for setattr on objects finished..."
	wait_delete_completed

	sync; sync_all_data || true

	echo "Verify disk usage after write"
	USED=$(getquota -u $TSTID global curspace)
	[ $USED -lt $TOTAL_BLKS ] &&
		error "Used space for user $TSTID:$USED, expected:$TOTAL_BLKS"
	USED=$(getquota -g $TSTID global curspace)
	[ $USED -lt $TOTAL_BLKS ] &&
		error "Used space for group $TSTID:$USED, expected:$TOTAL_BLKS"
	if is_project_quota_supported; then
		USED=$(getquota -p $TSTPRJID global curspace)
		[ $USED -lt $TOTAL_BLKS ] && error \
			"Used space for project $TSTPRJID:$USED, expected:$TOTAL_BLKS"
	fi

	echo "Verify inode usage after write"
	USED=$(getquota -u $TSTID global curinodes)
	[ $USED -lt $INODES ] &&
		error "Used inode for user $TSTID is $USED, expected $INODES"
	USED=$(getquota -g $TSTID global curinodes)
	[ $USED -lt $INODES ] &&
		error "Used inode for group $TSTID is $USED, expected $INODES"
	if is_project_quota_supported; then
		USED=$(getquota -p $TSTPRJID global curinodes)
		[ $USED -lt $INODES ] && error \
			"Used inode for project $TSTPRJID is $USED, expected $INODES"
	fi

	cleanup_quota_test

	echo "Verify disk usage after delete"
	USED=$(getquota -u $TSTID global curspace)
	[ $USED -eq 0 ] || error "Used space for user $TSTID isn't 0. $USED"
	USED=$(getquota -u $TSTID global curinodes)
	[ $USED -eq 0 ] || error "Used inodes for user $TSTID isn't 0. $USED"
	USED=$(getquota -g $TSTID global curspace)
	[ $USED -eq 0 ] || error "Used space for group $TSTID isn't 0. $USED"
	USED=$(getquota -g $TSTID global curinodes)
	[ $USED -eq 0 ] || error "Used inodes for group $TSTID isn't 0. $USED"
	if is_project_quota_supported; then
		USED=$(getquota -p $TSTPRJID global curspace)
		[ $USED -eq 0 ] ||
			error "Used space for project $TSTPRJID isn't 0. $USED"
		USED=$(getquota -p $TSTPRJID global curinodes)
		[ $USED -eq 0 ] ||
			error "Used inodes for project $TSTPRJID isn't 0. $USED"
	fi
}
run_test 33 "Basic usage tracking for user & group & project"

# usage transfer test for user & group & project
test_34() {
	local BLK_CNT=2 # 2MB
	local project_supported="no"

	is_project_quota_supported && project_supported="yes"
	setup_quota_test || error "setup quota failed with $?"

	# make sure the system is clean
	local USED=$(getquota -u $TSTID global curspace)
	[ $USED -ne 0 ] && error "Used space ($USED) for user $TSTID isn't 0."
	USED=$(getquota -g $TSTID global curspace)
	[ $USED -ne 0 ] && error "Used space ($USED) for group $TSTID isn't 0."

	local USED=$(getquota -u $TSTID2 global curspace)
	[ $USED -ne 0 ] && error "Used space ($USED) for user $TSTID2 isn't 0."
	if [ $project_supported == "yes" ]; then
		USED=$(getquota -p $TSTPRJID global curspace)
		[ $USED -ne 0 ] && error \
			"Used space ($USED) for Project $TSTPRJID isn't 0."
	fi

	echo "Write file..."
	$DD of=$DIR/$tdir/$tfile count=$BLK_CNT 2>/dev/null ||
		error "write failed"
	cancel_lru_locks osc
	sync; sync_all_data || true

	echo "chown the file to user $TSTID"
	chown $TSTID $DIR/$tdir/$tfile || error "chown failed"

	echo "Wait for setattr on objects finished..."
	wait_delete_completed

	BLK_CNT=$((BLK_CNT * 1024))

	echo "Verify disk usage for user $TSTID"
	USED=$(getquota -u $TSTID global curspace)
	[ $USED -lt $BLK_CNT ] &&
		error "Used space for user $TSTID is ${USED}, expected $BLK_CNT"
	USED=$(getquota -u $TSTID global curinodes)
	[ $USED -ne 1 ] &&
		error "Used inodes for user $TSTID is $USED, expected 1"

	echo "chgrp the file to group $TSTID"
	chgrp $TSTID $DIR/$tdir/$tfile || error "chgrp failed"

	echo "Wait for setattr on objects finished..."
	wait_delete_completed

	echo "Verify disk usage for group $TSTID"
	USED=$(getquota -g $TSTID global curspace)
	[ $USED -ge $BLK_CNT ] ||
		error "Used space for group $TSTID is $USED, expected $BLK_CNT"
	USED=$(getquota -g $TSTID global curinodes)
	[ $USED -eq 1 ] ||
		error "Used inodes for group $TSTID is $USED, expected 1"

	# chown won't change the ost object group. LU-4345 */
	echo "chown the file to user $TSTID2"
	chown $TSTID2 $DIR/$tdir/$tfile || error "chown to $TSTID2 failed"

	echo "Wait for setattr on objects finished..."
	wait_delete_completed

	echo "change_project project id to $TSTPRJID"
	[ $project_supported == "yes" ] &&
		change_project -p $TSTPRJID $DIR/$tdir/$tfile
	echo "Wait for setattr on objects finished..."
	wait_delete_completed

	echo "Verify disk usage for user $TSTID2/$TSTID and group $TSTID"
	USED=$(getquota -u $TSTID2 global curspace)
	[ $USED -lt $BLK_CNT ] &&
		error "Used space for user $TSTID2 is $USED, expected $BLK_CNT"
	USED=$(getquota -u $TSTID global curspace)
	[ $USED -ne 0 ] &&
		error "Used space for user $TSTID is $USED, expected 0"
	USED=$(getquota -g $TSTID global curspace)
	[ $USED -lt $BLK_CNT ] &&
		error "Used space for group $TSTID is $USED, expected $BLK_CNT"
	if [ $project_supported == "yes" ]; then
		USED=$(getquota -p $TSTPRJID global curspace)
		[ $USED -lt $BLK_CNT ] && error \
			"Used space for group $TSTPRJID is $USED, expected $BLK_CNT"
	fi
	return 0
}
run_test 34 "Usage transfer for user & group & project"

# usage is still accessible across restart
test_35() {
	local BLK_CNT=2 # 2 MB

	setup_quota_test || error "setup quota failed with $?"

	echo "Write file..."
	$RUNAS $DD of=$DIR/$tdir/$tfile count=$BLK_CNT 2>/dev/null ||
		error "write failed"
	is_project_quota_supported &&
		change_project -p $TSTPRJID $DIR/$tdir/$tfile
	cancel_lru_locks mdc
	cancel_lru_locks osc

	echo "Wait for setattr on objects finished..."
	wait_delete_completed

	sync; sync_all_data || true

	echo "Save disk usage before restart"
	local ORIG_USR_SPACE=$(getquota -u $TSTID global curspace)
	[ $ORIG_USR_SPACE -eq 0 ] &&
		error "Used space for user $TSTID is 0, expected ${BLK_CNT}M"
	local ORIG_USR_INODES=$(getquota -u $TSTID global curinodes)
	[ $ORIG_USR_INODES -eq 0 ] &&
		error "Used inodes for user $TSTID is 0, expected 1"
	echo "User $TSTID: ${ORIG_USR_SPACE}KB $ORIG_USR_INODES inodes"
	local ORIG_GRP_SPACE=$(getquota -g $TSTID global curspace)
	[ $ORIG_GRP_SPACE -eq 0 ] &&
		error "Used space for group $TSTID is 0, expected ${BLK_CNT}M"
	local ORIG_GRP_INODES=$(getquota -g $TSTID global curinodes)
	[ $ORIG_GRP_INODES -eq 0 ] &&
		error "Used inodes for group $TSTID is 0, expected 1"
	echo "Group $TSTID: ${ORIG_GRP_SPACE}KB $ORIG_GRP_INODES inodes"

	if is_project_quota_supported; then
		local ORIG_PRJ_SPACE=$(getquota -p $TSTPRJID global curspace)
		[ $ORIG_PRJ_SPACE -eq 0 ] && error \
			"Used space for project $TSTPRJID is 0, expected ${BLK_CNT}M"
		local ORIG_PRJ_INODES=$(getquota -p $TSTPRJID global curinodes)
		[ $ORIG_PRJ_INODES -eq 0 ] && error \
			"Used inodes for project $TSTPRJID is 0, expected 1"
		echo "Project $TSTPRJID: ${ORIG_PRJ_SPACE}KB $ORIG_PRJ_INODES inodes"
	fi

	log "Restart..."
	stopall
	setupall
	wait_recovery_complete
	quota_init

	echo "Verify disk usage after restart"
	local USED=$(getquota -u $TSTID global curspace)
	[ $USED -eq $ORIG_USR_SPACE ] || {
		ls -al $DIR/$tdir/$tfile
		$LFS quota -v -u $TSTID $DIR
		error "Used space for user $TSTID changed from " \
			"$ORIG_USR_SPACE to $USED"
	}
	USED=$(getquota -u $TSTID global curinodes)
	[ $USED -eq $ORIG_USR_INODES ] ||
		error "Used inodes for user $TSTID changed from " \
			"$ORIG_USR_INODES to $USED"
	USED=$(getquota -g $TSTID global curspace)
	[ $USED -eq $ORIG_GRP_SPACE ] ||
		error "Used space for group $TSTID changed from " \
			"$ORIG_GRP_SPACE to $USED"
	USED=$(getquota -g $TSTID global curinodes)
	[ $USED -eq $ORIG_GRP_INODES ] ||
		error "Used inodes for group $TSTID changed from " \
			"$ORIG_GRP_INODES to $USED"
	if [ $project_supported = "yes" ]; then
		USED=$(getquota -p $TSTPRJID global curinodes)
		[ $USED -eq $ORIG_PRJ_INODES ] ||
			error "Used inodes for project $TSTPRJID " \
				"changed from $ORIG_PRJ_INODES to $USED"
		USED=$(getquota -p $TSTPRJID global curspace)
		[ $USED -eq $ORIG_PRJ_SPACE ] ||
			error "Used space for project $TSTPRJID "\
				"changed from $ORIG_PRJ_SPACE to $USED"
	fi

	# check if the vfs_dq_init() is called before writing
	echo "Append to the same file..."
	$RUNAS $DD of=$DIR/$tdir/$tfile count=$BLK_CNT seek=1 2>/dev/null ||
		error "write failed"
	cancel_lru_locks osc
	sync; sync_all_data || true

	echo "Verify space usage is increased"
	USED=$(getquota -u $TSTID global curspace)
	[ $USED -gt $ORIG_USR_SPACE ] ||
		error "Used space for user $TSTID isn't increased" \
			"orig:$ORIG_USR_SPACE, now:$USED"
	USED=$(getquota -g $TSTID global curspace)
	[ $USED -gt $ORIG_GRP_SPACE ] ||
		error "Used space for group $TSTID isn't increased" \
			"orig:$ORIG_GRP_SPACE, now:$USED"
	if [ $project_supported = "yes" ]; then
		USED=$(getquota -p $TSTPRJID global curspace)
		[ $USED -gt $ORIG_PRJ_SPACE ] ||
			error "Used space for project $TSTPRJID isn't " \
				"increased orig:$ORIG_PRJ_SPACE, now:$USED"
	fi
}
run_test 35 "Usage is still accessible across reboot"

# chown/chgrp to the file created with MDS_OPEN_DELAY_CREATE
# LU-5006
test_37() {
	[ $(lustre_version_code $SINGLEMDS) -lt $(version_code 2.6.93) ] &&
		skip "Old server doesn't have LU-5006 fix."

	setup_quota_test || error "setup quota failed with $?"

	# make sure the system is clean
	local USED=$(getquota -u $TSTID global curspace)
	[ $USED -ne 0 ] &&
		error "Used space ($USED) for user $TSTID isn't 0."

	# create file with MDS_OPEN_DELAY_CREATE flag
	$SETSTRIPE -c 1 -i 0 $DIR/$tdir/$tfile ||
		error "Create file failed"
	# write to file
	dd if=/dev/zero of=$DIR/$tdir/$tfile bs=1M count=1 conv=notrunc \
		oflag=sync || error "Write file failed"
	# chown to the file
	chown $TSTID $DIR/$tdir/$tfile || error "Chown to file failed"

	# wait for setattr on objects finished..."
	wait_delete_completed

	USED=$(getquota -u $TSTID global curspace)
	[ $USED -ne 0 ] || quota_error u $TSTUSR "Used space is 0"
}
run_test 37 "Quota accounted properly for file created by 'lfs setstripe'"

# LU-8801
test_38() {
	[ $(lustre_version_code $SINGLEMDS) -lt $(version_code 2.8.60) ] &&
		skip "Old server doesn't have LU-8801 fix."

	[ "$UID" != 0 ] && skip_env "must run as root" && return

	setup_quota_test || error "setup quota failed with $?"

	# make sure the system is clean
	local USED=$(getquota -u $TSTID global curspace)
	[ $USED -ne 0 ] &&
		error "Used space ($USED) for user $TSTID isn't 0."
	USED=$(getquota -u $TSTID2 global curspace)
	[ $USED -ne 0 ] &&
		error "Used space ($USED) for user $TSTID2 isn't 0."

	local TESTFILE="$DIR/$tdir/$tfile"
	local file_cnt=10000

	# Generate id entries in accounting file
	echo "Create $file_cnt files..."
	for i in `seq $file_cnt`; do
		touch $TESTFILE-$i
		chown $((file_cnt - i)):$((file_cnt - i)) $TESTFILE-$i ||
			error "failed to chown $TESTFILE-$i"
	done
	cancel_lru_locks osc
	sync; sync_all_data || true

	local procf="osd-$(facet_fstype $SINGLEMDS).$FSNAME-MDT0000"
	procf=${procf}.quota_slave.acct_user
	local accnt_cnt

	acct_cnt=$(do_facet mds1 $LCTL get_param $procf | grep "id:" | \
		   awk '{if ($3 < 10000) {print $3}}' | wc -l)
	echo "Found $acct_cnt id entries"

	[ $file_cnt -eq $acct_cnt ] || {
		do_facet mds1 $LCTL get_param $procf
		error "skipped id entries"
	}
}
run_test 38 "Quota accounting iterator doesn't skip id entries"

test_39() {
	local TESTFILE="$DIR/$tdir/project"
	! is_project_quota_supported &&
		skip "Project quota is not supported"

	setup_quota_test || error "setup quota failed with $?"

	touch $TESTFILE
	projectid=$(lfs project $TESTFILE | awk '{print $1}')
	[ $projectid -ne 0 ] &&
		error "Project id should be 0 not $projectid"
	change_project -p 1024 $TESTFILE
	projectid=$(lfs project $TESTFILE | awk '{print $1}')
	[ $projectid -ne 1024 ] &&
		error "Project id should be 1024 not $projectid"

	stopall || error "failed to stopall (1)"
	mount
	setupall
	projectid=$(lfs project $TESTFILE | awk '{print $1}')
	[ $projectid -eq 1024 ] ||
		error "Project id should be 1024 not $projectid"
}
run_test 39 "Project ID interface works correctly"

test_40a() {
	! is_project_quota_supported &&
		skip "Project quota is not supported"
	local dir1="$DIR/$tdir/dir1"
	local dir2="$DIR/$tdir/dir2"

	setup_quota_test || error "setup quota failed with $?"

	mkdir -p $dir1 $dir2
	change_project -sp 1 $dir1 && touch $dir1/1
	change_project -sp 2 $dir2

	ln $dir1/1 $dir2/1_link &&
		error "Hard link across different project quota should fail"
	return 0
}
run_test 40a "Hard link across different project ID"

test_40b() {
	! is_project_quota_supported &&
		skip "Project quota is not supported"
	local dir1="$DIR/$tdir/dir1"
	local dir2="$DIR/$tdir/dir2"

	setup_quota_test || error "setup quota failed with $?"
	mkdir -p $dir1 $dir2
	change_project -sp 1 $dir1 && touch $dir1/1
	change_project -sp 2 $dir2

	mv $dir1/1 $dir2/2 || error "mv failed $?"
	local projid=$(lfs project $dir2/2 | awk '{print $1}')
	[ "$projid" -eq 2 ] || error "project id expected 2 not $projid"
}
run_test 40b "Mv across different project ID"

test_40c() {
	[ "$MDSCOUNT" -lt "2" ] && skip "needs >= 2 MDTs"
		! is_project_quota_supported &&
			skip "Project quota is not supported"

	setup_quota_test || error "setup quota failed with $?"
	local dir="$DIR/$tdir/dir"

	mkdir -p $dir && change_project -sp 1 $dir
	$LFS mkdir -i 1 $dir/remote_dir || error "create remote dir failed"
	local projid=$(lfs project -d $dir/remote_dir | awk '{print $1}')
	[ "$projid" != "1" ] && error "projid id expected 1 not $projid"
	touch $dir/remote_dir/file
	#verify inherit works file for remote dir.
	local projid=$(lfs project -d $dir/remote_dir/file | awk '{print $1}')
	[ "$projid" != "1" ] &&
		error "file under remote dir expected 1 not $projid"

	#Agent inode should be ignored for project quota
	local used=$(getquota -p 1 global curinodes)
	[ $used -eq 3 ] ||
		error "file count expected 3 got $used"
}
run_test 40c "Remote child Dir inherit project quota properly"

test_40d() {
	[ "$MDSCOUNT" -lt "2" ] && skip_env "needs >= 2 MDTs"
	is_project_quota_supported || skip "Project quota is not supported"

	setup_quota_test || error "setup quota failed with $?"
	local dir="$DIR/$tdir/dir"

	mkdir -p $dir
	$LFS setdirstripe -D -c 2 -i -1 $dir || error "setdirstripe failed"
	change_project -sp $TSTPRJID $dir ||
		error "change project on $dir failed"
	for i in $(seq 5); do
		mkdir -p $dir/d$i/d$i ||
			error "mkdir $dir/d$i/d$i failed"
		local projid=$($LFS project -d $dir/d$i/d$i |
			       awk '{print $1}')
		[ "$projid" == "$TSTPRJID" ] ||
			error "projid id expected $TSTPRJID not $projid"
		touch $dir/d$i/d$i/file
		#verify inherit works file for stripe dir.
		local projid=$($LFS project -d $dir/d$i/d$i/file | awk '{print $1}')
		[ "$projid" == "$TSTPRJID" ] ||
			error "file under remote dir expected 1 not $projid"
	done

	# account should be 1 + (2 + 1) *10 + 1 * 5
	local used=$(getquota -p $TSTPRJID global curinodes)
	[ $used -eq 36 ] ||
		error "file count expected 36 got $used"
}
run_test 40d "Stripe Directory inherit project quota properly"

test_41() {
	is_project_quota_supported ||
		skip "Project quota is not supported"
	setup_quota_test || error "setup quota failed with $?"
	local dir="$DIR/$tdir/dir"
	local blimit=102400
	local ilimit=4096
	local projid=$((testnum * 1000))

	quota_init

	# enable mdt/ost quota
	set_mdt_qtype ugp || error "enable mdt quota failed"
	set_ost_qtype ugp || error "enable ost quota failed"

	test_mkdir -p $dir && change_project -sp $projid $dir
	$LFS setquota -p $projid -b 0 -B ${blimit}K -i 0 -I $ilimit $dir ||
		error "set project quota failed"

	sync; sync_all_data
	sleep_maxage

	# check if df output works as expected
	echo "== global statfs: $MOUNT =="
	df -kP $MOUNT; df -iP $MOUNT; $LFS quota -p $projid $dir
	echo
	echo "== project statfs (prjid=$projid): $dir =="
	df -kP $dir; df -iP $dir
	local bused=$(getquota -p $projid global curspace)
	local iused=$(getquota -p $projid global curinodes)
	# note trailing space to match double printf from awk
	local expected="$blimit $bused $ilimit $iused "

	wait_update $HOSTNAME \
		"{ df -kP $dir; df -iP $dir; } |
		 awk '/$FSNAME/ { printf \\\"%d %d \\\", \\\$2,\\\$3 }'" \
		"$expected" ||
		error "failed to get correct statfs for project quota"
}
run_test 41 "df should return projid-specific values"

test_delete_qid()
{
	local qslv_file=$1
	local qtype_file=$2
	local qtype=$3
	local qid=$4
	local osd="osd-ldiskfs"

	[ "$ost1_FSTYPE" = zfs ] && osd="osd-zfs"

	rm -f $DIR/$tdir/$tfile
	$LFS setstripe -i 0 -c 1 $DIR/$tdir/$tfile
	chmod a+rw $DIR/$tdir/$tfile

	$LFS setquota $qtype $qid -B 300M $MOUNT
	$RUNAS dd if=/dev/zero of=$DIR/$tdir/$tfile bs=1M count=1 ||
		error "failed to dd"

	do_facet $SINGLEMDS \
		"cat /proc/fs/lustre/qmt/$FSNAME-QMT0000/dt-0x0/$qtype_file |
		 grep -E 'id: *$qid'" || error "QMT: no qid $qid is found"
	echo $osd
	do_facet ost1 \
		"cat /proc/fs/lustre/$osd/$FSNAME-OST0000/$qslv_file |
		 grep -E 'id: *$qid'" || error "QSD: no qid $qid is found"

	$LFS setquota $qtype $qid --delete $MOUNT
	do_facet $SINGLEMDS \
		"cat /proc/fs/lustre/qmt/$FSNAME-QMT0000/dt-0x0/$qtype_file |
		 grep -E 'id: *$qid'" && error "QMT: qid $qid is not deleted"
	sleep 5
	do_facet ost1 \
		"cat /proc/fs/lustre/$osd/$FSNAME-OST0000/$qslv_file |
		 grep -E 'id: *$qid'" && error "QSD: qid $qid is not deleted"

	$LFS setquota $qtype $qid -B 500M $MOUNT
	$RUNAS dd if=/dev/zero of=$DIR/$tdir/$tfile bs=1M count=1 ||
		error "failed to dd"
	do_facet $SINGLEMDS \
		"cat /proc/fs/lustre/qmt/$FSNAME-QMT0000/dt-0x0/$qtype_file |
		 grep -E 'id: *$qid'" || error "QMT: qid $pid is not recreated"
	cat /proc/fs/lustre/$osd/$FSNAME-OST0000/$qslv_file
	do_facet ost1 \
		"cat /proc/fs/lustre/$osd/$FSNAME-OST0000/$qslv_file |
		 grep -E 'id: *$qid'" || error "QSD: qid $qid is not recreated"
}

test_48()
{
	(( "$MDS1_VERSION" >= $(version_code 2.12.6.72) )) ||
		skip "Need MDS >= 2.12.6.72 to delete quota"
	setup_quota_test || error "setup quota failed with $?"
	set_ost_qtype $QTYPE || error "enable ost quota failed"
	quota_init

	test_delete_qid "quota_slave/limit_user" "glb-usr" "-u" $TSTID
	test_delete_qid "quota_slave/limit_group" "glb-grp" "-g" $TSTID
	is_project_quota_supported &&
	    test_delete_qid "quota_slave/limit_project" "glb-prj" "-p" "10000"

}
run_test 48 "lfs quota --delete should delete quota project ID"

test_50() {
	! is_project_quota_supported &&
		skip "Project quota is not supported"

	setup_quota_test || error "setup quota failed with $?"
	local dir1="$DIR/$tdir/dir1"
	local dir2="$DIR/$tdir/dir2"

	mkdir -p $dir1 && change_project -sp 1 $dir1
	mkdir -p $dir2 && change_project -sp 2 $dir2
	for num in $(seq 1 10); do
		touch $dir1/file_$num $dir2/file_$num
		ln -s $dir1/file_$num $dir1/file_$num"_link"
		ln -s $dir2/file_$num $dir2/file_$num"_link"
	done

	count=$($LFS find --projid 1 $DIR | wc -l)
	[ "$count" != 21 ] && error "expected 21 but got $count"

	# 1(projid 0 dir) + 1(projid 2 dir) + 20(projid 2 files)
	count=$($LFS find ! --projid 1 $DIR/$tdir | wc -l)
	[ $count -eq 22 ] || error "expected 22 but got $count"
}
run_test 50 "Test if lfs find --projid works"

test_51() {
	! is_project_quota_supported &&
		skip "Project quota is not supported"
	setup_quota_test || error "setup quota failed with $?"
	local dir="$DIR/$tdir/dir"

	mkdir $dir && change_project -sp 1 $dir
	local used=$(getquota -p 1 global curinodes)
	[ $used != "1" ] && error "expected 1 got $used"

	touch $dir/1
	touch $dir/2
	cp $dir/2 $dir/3
	used=$(getquota -p 1 global curinodes)
	[ $used != "4" ] && error "expected 4 got $used"

	$DD if=/dev/zero of=$DIR/$tdir/6 bs=1M count=1
	#try cp to dir
	cp $DIR/$tdir/6 $dir/6
	used=$(getquota -p 1 global curinodes)
	[ $used != "5" ] && error "expected 5 got $used"

	#try mv to dir
	mv $DIR/$tdir/6 $dir/7
	used=$(getquota -p 1 global curinodes)
	[ $used -eq 6 ] || error "expected 6 got $used"
}
run_test 51 "Test project accounting with mv/cp"

test_52() {
	! is_project_quota_supported &&
		skip "Project quota is not supported"

	(( MDS1_VERSION >= $(version_code 2.12.9.4) )) ||
		skip "Need MDS version at least 2.12.9.4"

	setup_quota_test || error "setup quota failed with $?"

	local dir1=$DIR/$tdir/t52_dir1
	local dir2=$DIR/$tdir/t52_dir2

	mkdir $dir1 || error "failed to mkdir $dir1"
	mkdir $dir2 || error "failed to mkdir $dir2"

	$LFS project -sp 1000 $dir1 || error "fail to set project on $dir1"
	$LFS project -sp 1001 $dir2 || error "fail to set project on $dir2"

	stack_trap "rm -f $dir1/$tfile $dir2/$tfile"
	$DD if=/dev/zero of=$dir1/$tfile bs=1M count=100 ||
		error "failed to create and write $tdir1/$tfile"

	cancel_lru_locks osc
	sync; sync_all_data || true

	local attrs=($(lsattr -p $dir1/$tfile))
	(( ${attrs[0]} == 1000 )) ||
		error "project ID on $dir1/$tfile is not inherited"

	$LFS quota -p 1000 $DIR
	$LFS quota -p 1001 $DIR

	local prev_used=$(getquota -p 1000 global curspace)
	local prev_used2=$(getquota -p 1001 global curspace)

	mrename $dir1 $dir2/tdir || log "rename directory return $?"

	local inum_before=$(ls -i $dir1/$tfile | awk '{print $1}')
	mrename $dir1/$tfile $dir2/$tfile || error "failed to rename file"
	local inum_after=$(ls -i $dir2/$tfile | awk '{print $1}')

	attrs=($(lsattr -p $dir2/$tfile))
	(( ${attrs[0]} == 1001 )) ||
		error "project ID is not updated after rename"

	(( $inum_before == $inum_after )) ||
		error "inode is changed after rename: $inum_before, $inum_after"

	sync_all_data || true

	$LFS quota -p 1000 $DIR
	$LFS quota -p 1001 $DIR

	local new_used=$(getquota -p 1000 global curspace)
	local new_used2=$(getquota -p 1001 global curspace)

	(( $prev_used >= $new_used + 102400 )) ||
		error "quota is not deducted from old project ID"
	(( $prev_used2 <= $new_used2 - 102400 )) ||
		error "quota is not added for the new project ID"
}
run_test 52 "Rename normal file across project ID"

test_53() {
	! is_project_quota_supported &&
		skip "Project quota is not supported"
	setup_quota_test || error "setup quota failed with $?"
	local dir="$DIR/$tdir/dir"
	mkdir $dir && change_project -s $dir
	[[ $($LFS project -d $dir) =~ " P " ]] ||
		error "inherit attribute should be set"

	change_project -C $dir
	[[ $($LFS project -d $dir) =~ " - " ]] ||
		error "inherit attribute should be cleared"
}
run_test 53 "Project inherit attribute could be cleared"

test_54() {
	! is_project_quota_supported &&
		skip "Project quota is not supported"
	setup_quota_test || error "setup quota failed with $?"
	local testfile="$DIR/$tdir/$tfile-0"

	#set project ID/inherit attribute
	change_project -sp $TSTPRJID $DIR/$tdir
	$RUNAS createmany -m ${testfile} 100 ||
		error "create many files failed"

	local proj_count=$(lfs project -r $DIR/$tdir | wc -l)
	# one more count for directory itself */
	((proj_count++))

	#check project
	local proj_count1=$(lfs project -rcp $TSTPRJID $DIR/$tdir | wc -l)
	[ $proj_count1 -eq 0 ] || error "c1: expected 0 got $proj_count1"

	proj_count1=$(lfs project -rcp $((TSTPRJID+1)) $DIR/$tdir | wc -l)
	[ $proj_count1 -eq $proj_count ] ||
			error "c2: expected $proj_count got $proj_count1"

	#clear project but with kept projid
	change_project -rCk $DIR/$tdir
	proj_count1=$(lfs project -rcp $TSTPRJID $DIR/$tdir | wc -l)
	[ $proj_count1 -eq 1 ] ||
			error "c3: expected 1 got $proj_count1"

	#verify projid untouched.
	proj_count1=$(lfs project -r $DIR/$tdir | grep -c $TSTPRJID)
	((proj_count1++))
	[ $proj_count1 -eq $proj_count ] ||
			error "c4: expected $proj_count got $proj_count1"

	# test -0 option
	lfs project $DIR/$tdir -cr -0 | xargs -0 lfs project -s
	proj_count1=$(lfs project -rcp $TSTPRJID $DIR/$tdir | wc -l)
	[ $proj_count1 -eq 0 ] || error "c5: expected 0 got $proj_count1"

	#this time clear all
	change_project -rC $DIR/$tdir
	proj_count1=$(lfs project -r $DIR/$tdir | grep -c $TSTPRJID)
	[ $proj_count1 -eq 0 ] ||
			error "c6: expected 0 got $proj_count1"
	#cleanup
	unlinkmany ${testfile} 100 ||
		error "unlink many files failed"
}
run_test 54 "basic lfs project interface test"

test_55() {
	[ $(lustre_version_code $SINGLEMDS) -lt $(version_code 2.10.58) ] &&
		skip "Not supported before 2.10.58."
	setup_quota_test || error "setup quota failed with $?"

	set_ost_qtype $QTYPE || error "enable ost quota failed"
	quota_init

	#add second group to TSTUSR
	usermod -G $TSTUSR,$TSTUSR2 $TSTUSR

	#prepare test file
	$RUNAS dd if=/dev/zero of=$DIR/$tdir/$tfile bs=1024 count=100000 ||
	error "failed to dd"

	cancel_lru_locks osc
	sync; sync_all_data || true

	$LFS setquota -g $TSTUSR2 -b 0 -B 50M $DIR ||
	error "failed to setquota on group $TSTUSR2"

	$LFS quota -v -g $TSTUSR2 $DIR

	runas -u $TSTUSR -g $TSTUSR2 chgrp $TSTUSR2 $DIR/$tdir/$tfile &&
	error "chgrp should failed with -EDQUOT"

	USED=$(getquota -g $TSTUSR2 global curspace)
	echo "$USED"

	$LFS setquota -g $TSTUSR2 -b 0 -B 300M $DIR ||
	error "failed to setquota on group $TSTUSR2"

	$LFS quota -v -g $TSTUSR2 $DIR

	runas -u $TSTUSR -g $TSTUSR2 chgrp $TSTUSR2 $DIR/$tdir/$tfile ||
	error "chgrp should succeed"

	$LFS quota -v -g $TSTUSR2 $DIR

}
run_test 55 "Chgrp should be affected by group quota"

test_56() {
	setup_quota_test || error "setup quota failed with $?"

	set_ost_qtype $QTYPE || error "enable ost quota failed"
	quota_init

	$LFS setquota -t -u -b 10 -i 10 $DIR ||
		erro "failed to set grace time for usr quota"
	grace_time=$($LFS quota -t -u $DIR | grep "Block grace time:" |
		     awk '{print $4 $8}')
	if [ "x$grace_time" != "x10s;10s" ]; then
		$LFS quota -t -u $DIR
		error "expected grace time: 10s;10s, got:$grace_time"
	fi
}
run_test 56 "lfs quota -t should work well"

test_57() {
	setup_quota_test || error "setup quota failed with $?"

	local dir="$DIR/$tdir/dir"
	mkdir -p $dir
	mkfifo $dir/pipe
	#command can process further if it hit some errors
	$LFS project -sp 1 $dir/pipe
	touch $dir/aaa $dir/bbb
	mkdir $dir/subdir -p
	touch $dir/subdir/aaa $dir/subdir/bbb
	#create one invalid link file
	ln -s $dir/not_exist_file $dir/ccc
	local cnt=$(lfs project -r $dir 2>/dev/null | wc -l)
	[ $cnt -eq 7 ] || error "expected 7 got $cnt"
}
run_test 57 "lfs project could tolerate errors"

# LU-16988
test_mirror()
{
	local projid=$1
	local testfile=$2
	local mirrorfile=$3

	# create mirror
	$LFS mirror extend -N2 $mirrorfile || error "failed to create mirror"

	local mirrors=$($LFS getstripe -N $testfile)
	[[ $mirrors == 3 ]] || error "mirror count $mirrors is wrong"

	cancel_lru_locks osc
	cancel_lru_locks mdc
	sync; sync_all_data || true

	local prev_usage=$(getquota -p $projid global curspace)

	$RUNAS $DD of=$testfile count=50 conv=nocreat oflag=direct ||
			quota_error p $projid "write failed, expect succeed"

	cancel_lru_locks osc
	cancel_lru_locks mdc
	sync; sync_all_data || true

	$RUNAS $LFS mirror resync $testfile || error "failed to resync mirror"

	local usage=$(getquota -p $projid global curspace)
	(( usage >= prev_usage + 150*1024 )) ||
				error "project quota $usage is wrong"

	$RUNAS $DD of=$testfile count=30 conv=nocreat seek=50 oflag=direct ||
			quota_error p $projid "write failed, expect succeed"

	$RUNAS $LFS mirror resync $testfile &&
			error "resync mirror succeed, expect EDQUOT"

	$LFS mirror split -d --mirror-id 2 $testfile ||
			error "failed to delete the second mirror"
	$LFS mirror split -d --mirror-id 3 $testfile ||
			error "failed to delete the third mirror"
}

test_58() {
	(( $MDS1_VERSION >= $(version_code 2.12.9.31) )) ||
		skip "need MDS 2.12.9.31 or later"

	is_project_quota_supported || skip "Project quota is not supported"

	local testdir="$DIR/$tdir"
	local testfile="$DIR/$tdir/$tfile"
	local projid=1000
	local projid2=1001

	setup_quota_test || error "setup quota failed with $?"

	USED=$(getquota -p $projid global curspace)
	[ $USED -ne 0 ] && error "Used space ($USED) for proj $projid isn't 0"

	USED=$(getquota -p $projid2 global curspace)
	[ $USED -ne 0 ] && error "Used space ($USED) for proj $projid2 isn't 0"

	chown $TSTUSR.$TSTUSR $testdir || error "chown $testdir failed"
	quota_init
	set_ost_qtype ugp || error "enable ost quota failed"

	$LFS project -sp $projid $testdir || error "failed to set project ID"
	$LFS setquota -p $projid -B 200M $DIR ||
				error "failed to to set prj $projid quota"

	$RUNAS touch $testfile

	local id=$(lfs project -d $testfile | awk '{print $1}')
	[ "$id" != "$projid" ] && error "projid $projid is not inherited $id"

	echo "test by mirror created with normal file"
	test_mirror $projid $testfile $testfile

	$TRUNCATE $testfile 0
	wait_delete_completed || error "wait_delete_completed failed"
	sync_all_data || true

	$LFS project -sp $projid2 $testdir ||
				error "failed to set directory project ID"
	$LFS project -p $projid2 $testfile ||
				error "failed to set file project ID"
	$LFS setquota -p $projid -b 0 -B 0 $DIR ||
				error "failed to to reset prj quota"
	$LFS setquota -p $projid2 -B 200M $DIR ||
				error "failed to to set prj $projid2 quota"

	local fid=$($LFS path2fid $testfile)

	echo "test by mirror created with FID"
	test_mirror $projid2 $testfile $MOUNT/.lustre/fid/$fid
}
run_test 58 "project ID should be kept for new mirrors created by FID"

test_59() {
	[ "$(facet_fstype $SINGLEMDS)" != "ldiskfs" ] &&
		skip "ldiskfs only test"
	disable_project_quota
	setup_quota_test || error "setup quota failed with $?"
	quota_init

	local testfile="$DIR/$tdir/$tfile-0"
	#make sure it did not crash kernel
	touch $testfile && lfs project -sp 1 $testfile

	enable_project_quota
}
run_test 59 "lfs project dosen't crash kernel with project disabled"

test_60() {
	[ $MDS1_VERSION -lt $(version_code 2.11.53) ] &&
		skip "Needs MDS version 2.11.53 or later."
	setup_quota_test || error "setup quota failed with $?"

	local testfile=$DIR/$tdir/$tfile
	local limit=100

	set_mdt_qtype "ug" || error "enable mdt quota failed"

	$LFS setquota -g $TSTUSR -b 0 -B 0 -i 0 -I $limit $DIR ||
		error "set quota failed"
	quota_show_check a g $TSTUSR

	chown $TSTUSR.$TSTUSR $DIR/$tdir || error "chown $DIR/$tdir failed"
	chmod g+s $DIR/$tdir || error "chmod g+s failed"
	$RUNAS createmany -m ${testfile} $((limit-1)) ||
		error "create many files failed"

	$RUNAS touch $DIR/$tdir/foo && error "regular user should fail"

	# root user can overrun quota
	runas -u 0 -g 0 touch $DIR/$tdir/foo ||
		error "root user should succeed"

}
run_test 60 "Test quota for root with setgid"

# test default quota
test_default_quota() {
	[ $(lustre_version_code $SINGLEMDS) -lt $(version_code 2.11.51) ] &&
		skip "Not supported before 2.11.51."

	local qtype=$1
	local qpool=$2
	local qid=$TSTUSR
	local qprjid=$TSTPRJID
	local qdtype="-U"
	local qs="-b"
	local qh="-B"
	local LIMIT=20480 #20M disk space
	local TESTFILE="$DIR/$tdir/$tfile-0"

	[ $qtype == "-p" ] && ! is_project_quota_supported &&
		echo "Project quota is not supported" && return 0

	[ $qtype == "-u" ] && qdtype="-U"
	[ $qtype == "-g" ] && qdtype="-G"
	[ $qtype == "-p" ] && {
		qdtype="-P"
		qid=$qprjid
	}

	[ $qpool == "meta" ] && {
		LIMIT=10240 #10K inodes
		qs="-i"
		qh="-I"
	}

	setup_quota_test || error "setup quota failed with $?"

	quota_init

	# enable mdt/ost quota
	set_mdt_qtype $QTYPE || error "enable mdt quota failed"
	set_ost_qtype $QTYPE || error "enable ost quota failed"

	log "set to use default quota"
	$LFS setquota $qtype $qid -d $DIR ||
		error "set $qid to use default quota failed"

	log "set default quota"
	$LFS setquota $qdtype $qs ${LIMIT} $qh ${LIMIT} $DIR ||
		error "set $qid default quota failed"

	log "get default quota"
	$LFS quota $qdtype $DIR || error "get default quota failed"

	if [ $qpool == "data" ]; then
		local SLIMIT=$($LFS quota $qdtype $DIR | grep "$MOUNT" | \
							awk '{print $2}')
		[ $SLIMIT -eq $LIMIT ] ||
			error "the returned default quota is wrong"
	else
		local SLIMIT=$($LFS quota $qdtype $DIR | grep "$MOUNT" | \
							awk '{print $5}')
		[ $SLIMIT -eq $LIMIT ] ||
			error "the returned default quota is wrong"
	fi

	# make sure the system is clean
	local USED=$(getquota $qtype $qid global curspace)
	[ $USED -ne 0 ] && error "Used space for $qid isn't 0."

	$SETSTRIPE $TESTFILE -c 1 || error "setstripe $TESTFILE failed"
	chown $TSTUSR.$TSTUSR $TESTFILE || error "chown $TESTFILE failed"

	[ $qtype == "-p" ] && change_project -sp $TSTPRJID $DIR/$tdir

	log "Test not out of quota"
	if [ $qpool == "data" ]; then
		$RUNAS $DD of=$TESTFILE count=$((LIMIT/2 >> 10)) oflag=sync ||
			quota_error $qtype $qid "write failed, expect succeed"
	else
		$RUNAS createmany -m $TESTFILE $((LIMIT/2)) ||
			quota_error $qtype $qid "create failed, expect succeed"

		unlinkmany $TESTFILE $((LIMIT/2))
	fi

	log "Test out of quota"
	# flush cache, ensure noquota flag is set on client
	cancel_lru_locks osc
	cancel_lru_locks mdc
	sync; sync_all_data || true
	if [ $qpool == "data" ]; then
		$RUNAS $DD of=$TESTFILE count=$((LIMIT*2 >> 10)) oflag=sync &&
			quota_error $qtype $qid "write succeed, expect EDQUOT"
	else
		$RUNAS createmany -m $TESTFILE $((LIMIT*2)) &&
			quota_error $qtype $qid "create succeed, expect EDQUOT"

		unlinkmany $TESTFILE $((LIMIT*2))
	fi

	log "Increase default quota"
	# increase default quota
	$LFS setquota $qdtype $qs $((LIMIT*3)) $qh $((LIMIT*3)) $DIR ||
		error "set default quota failed"

	cancel_lru_locks osc
	cancel_lru_locks mdc
	sync; sync_all_data || true
	if [ $qpool == "data" ]; then
		$RUNAS $DD of=$TESTFILE count=$((LIMIT*2 >> 10)) oflag=sync ||
			quota_error $qtype $qid "write failed, expect succeed"
	else
		$RUNAS createmany -m $TESTFILE $((LIMIT*2)) ||
			quota_error $qtype $qid "create failed, expect succeed"

		unlinkmany $TESTFILE $((LIMIT*2))
	fi

	log "Set quota to override default quota"
	$LFS setquota $qtype $qid $qs ${LIMIT} $qh ${LIMIT} $DIR ||
		error "set $qid quota failed"

	cancel_lru_locks osc
	cancel_lru_locks mdc
	sync; sync_all_data || true
	if [ $qpool == "data" ]; then
		$RUNAS $DD of=$TESTFILE count=$((LIMIT*2 >> 10)) oflag=sync &&
			quota_error $qtype $qid "write succeed, expect EQUOT"
	else
		$RUNAS createmany -m $TESTFILE $((LIMIT*2)) &&
			quota_error $qtype $qid "create succeed, expect EQUOT"

		unlinkmany $TESTFILE $((LIMIT*2))
	fi

	log "Set to use default quota again"
	$LFS setquota $qtype $qid -d $DIR ||
		error "set $qid to use default quota failed"

	cancel_lru_locks osc
	cancel_lru_locks mdc
	sync; sync_all_data || true
	if [ $qpool == "data" ]; then
		$RUNAS $DD of=$TESTFILE count=$((LIMIT*2 >> 10)) oflag=sync ||
			quota_error $qtype $qid "write failed, expect succeed"
	else
		$RUNAS createmany -m $TESTFILE $((LIMIT*2)) ||
			quota_error $qtype $qid "create failed, expect succeed"

		unlinkmany $TESTFILE $((LIMIT*2))
	fi

	log "Cleanup"
	rm -f $TESTFILE
	wait_delete_completed || error "wait_delete_completed failed"
	sync_all_data || true
	$LFS setquota $qdtype -b 0 -B 0 -i 0 -I 0 $DIR ||
		error "reset default quota failed"
	$LFS setquota $qtype $qid -b 0 -B 0 -i 0 -I 0 $DIR ||
		error "reset quota failed"
	cleanup_quota_test
}

test_61() {
	test_default_quota "-u" "data"
	test_default_quota "-u" "meta"
	test_default_quota "-g" "data"
	test_default_quota "-g" "meta"
	test_default_quota "-p" "data"
	test_default_quota "-p" "meta"
}
run_test 61 "default quota tests"

test_62() {
	! is_project_quota_supported &&
		skip "Project quota is not supported"
	 [[ "$(chattr -h 2>&1)" =~ "project" ]] ||
		skip "chattr did not support project quota"
	setup_quota_test || error "setup quota failed with $?"
	local testdir=$DIR/$tdir/

	$RUNAS mkdir -p $testdir || error "failed to mkdir"
	change_project -s $testdir
	[[ $($LFS project -d $testdir) =~ "P" ]] ||
		error "inherit attribute should be set"
	# chattr used FS_IOC_SETFLAGS ioctl
	$RUNAS chattr -P $testdir &&
		error "regular user clear inherit should fail"
	[[ $($LFS project -d $testdir) =~ "P" ]] ||
		error "inherit attribute should still be set"
	chattr -P $testdir || error "root failed to clear inherit"
	[[ $($LFS project -d $testdir) =~ "P" ]] &&
		error "inherit attribute should be cleared"
	return 0
}
run_test 62 "Project inherit should be only changed by root"

test_dom() {
	[ $(lustre_version_code $SINGLEMDS) -lt $(version_code 2.11.55) ] &&
		skip "Not supported before 2.11.55" && return

	local qtype=$1
	local qid=$TSTUSR
	local dd_failed=false
	local tdir_dom=${tdir}_dom
	local LIMIT=20480 #20M

	[ $qtype == "p" ] && ! is_project_quota_supported &&
		echo "Project quota is not supported" && return 0

	[ $qtype == "p" ] && qid=$TSTPRJID

	setup_quota_test || error "setup quota failed with $?"

	quota_init

	# enable mdt/ost quota
	set_mdt_qtype $QTYPE || error "enable mdt quota failed"
	set_ost_qtype $QTYPE || error "enable ost quota failed"

	# make sure the system is clean
	local USED=$(getquota -$qtype $qid global curspace)
	[ $USED -ne 0 ] && error "Used space for $qid isn't 0."

	chown $TSTUSR.$TSTUSR $DIR/$tdir || error "chown $tdir failed"

	mkdir $DIR/$tdir_dom || error "mkdir $tdir_dom failed"
	$SETSTRIPE -E 1M -L mdt $DIR/$tdir_dom ||
		error "setstripe $tdir_dom failed"
	chown $TSTUSR.$TSTUSR $DIR/$tdir_dom || error "chown $tdir_dom failed"

	[ $qtype == "p" ] && {
		change_project -sp $TSTPRJID $DIR/$tdir
		change_project -sp $TSTPRJID $DIR/$tdir_dom
	}

	$LFS setquota -$qtype $qid -b $LIMIT -B $LIMIT $DIR ||
		error "set $qid quota failed"

	for ((i = 0; i < $((LIMIT/2048)); i++)); do
		$RUNAS $DD of=$DIR/$tdir_dom/$tfile-$i count=1 oflag=sync ||
								dd_failed=true
	done

	$dd_failed && quota_error $qtype $qid "write failed, expect succeed"

	for ((i = $((LIMIT/2048)); i < $((LIMIT/1024 + 10)); i++)); do
		$RUNAS $DD of=$DIR/$tdir_dom/$tfile-$i count=1 oflag=sync ||
								dd_failed=true
	done

	$dd_failed || quota_error $qtype $qid "write succeed, expect EDQUOT"

	rm -f $DIR/$tdir_dom/*

	# flush cache, ensure noquota flag is set on client
	cancel_lru_locks osc
	cancel_lru_locks mdc
	sync; sync_all_data || true

	dd_failed=false

	$RUNAS $DD of=$DIR/$tdir/file count=$((LIMIT/2048)) oflag=sync ||
		quota_error $qtype $qid "write failed, expect succeed"

	for ((i = 0; i < $((LIMIT/2048 + 10)); i++)); do
		$RUNAS $DD of=$DIR/$tdir_dom/$tfile-$i count=1 oflag=sync ||
								dd_failed=true
	done

	$dd_failed || quota_error $qtype $TSTID "write succeed, expect EDQUOT"

	rm -f $DIR/$tdir/*
	rm -f $DIR/$tdir_dom/*

	# flush cache, ensure noquota flag is set on client
	cancel_lru_locks osc
	cancel_lru_locks mdc
	sync; sync_all_data || true

	dd_failed=false

	for ((i = 0; i < $((LIMIT/2048)); i++)); do
		$RUNAS $DD of=$DIR/$tdir_dom/$tfile-$i count=1 oflag=sync ||
								dd_failed=true
	done

	$dd_failed && quota_error $qtype $qid "write failed, expect succeed"

	$RUNAS $DD of=$DIR/$tdir/file count=$((LIMIT/2048 + 10)) oflag=sync &&
		quota_error $qtype $qid "write succeed, expect EDQUOT"

	rm -fr $DIR/$tdir
	rm -fr $DIR/$tdir_dom

	$LFS setquota -u $TSTUSR -b 0 -B 0 -i 0 -I 0 $DIR ||
		error "reset usr quota failed"
}

test_63() {
	test_dom "u"
	test_dom "g"
	test_dom "p"
}
run_test 63 "quota on DoM tests"

test_64() {
	! is_project_quota_supported &&
		skip "Project quota is not supported"
	setup_quota_test || error "setup quota failed with $?"
	local dir1="$DIR/$tdir/"

	touch $dir1/file
	ln -s $dir1/file $dir1/file_link
	mkfifo $dir1/fifo

	$LFS project -srp $TSTPRJID $dir1 >&/dev/null ||
		error "set project should succeed"

	used=$(getquota -p $TSTPRJID global curinodes)
	[ $used -eq 4 ] || error "expected 4 got $used"
	$LFS project -rC $dir1 >&/dev/null ||
		error "clear project should succeed"

	used=$(getquota -p $TSTPRJID global curinodes)
	[ $used -eq 0 ] || error "expected 0 got $used"
}
run_test 64 "lfs project on non dir/files should succeed"

test_65() {
	local SIZE=10 #10M
	local TESTFILE="$DIR/$tdir/$tfile-0"

	setup_quota_test || error "setup quota failed with $?"
	set_ost_qtype $QTYPE || error "enable ost quota failed"
	quota_init

	echo "Write..."
	$RUNAS $DD of=$TESTFILE count=$SIZE ||
		error "failed to write"
	# flush cache, ensure noquota flag is set on client
	cancel_lru_locks osc
	sync; sync_all_data || true

	local quota_u=$($LFS quota -u $TSTUSR $DIR)
	local quota_g=$($LFS quota -g $TSTUSR $DIR)
	local quota_all=$($RUNAS $LFS quota $DIR)

	[ "$(echo "$quota_all" | head -n3)" == "$quota_u" ] ||
		error "usr quota not match"
	[ "$(echo "$quota_all" | tail -n3)" == "$quota_g" ] ||
		error "grp quota not match"
}
run_test 65 "Check lfs quota result"

test_66() {
	! is_project_quota_supported &&
		skip "Project quota is not supported"
	[ "$MDS1_VERSION" -lt $(version_code 2.12.4) ] &&
		skip "Not supported before 2.12.4"
	setup_quota_test || error "setup quota failed with $?"
	local old=$(do_facet mds1 $LCTL get_param -n \
		    mdt.*.enable_chprojid_gid | head -1)
	local testdir=$DIR/$tdir/foo

	do_facet mds1 $LCTL set_param mdt.*.enable_chprojid_gid=0
	stack_trap "do_facet mds1 $LCTL set_param mdt.*.enable_chprojid_gid=0" \
		EXIT

	test_mkdir -i 0 -c 1 $testdir || error "failed to mkdir"
	chown -R $TSTID:$TSTID $testdir
	change_project -sp $TSTPRJID $testdir
	$RUNAS mkdir $testdir/foo || error "failed to mkdir foo"

	$RUNAS lfs project -p 0 $testdir/foo &&
		error "nonroot user should fail to set projid"

	$RUNAS lfs project -C $testdir/foo &&
		error "nonroot user should fail to clear projid"

	change_project -C $testdir/foo || error "failed to clear project"

	do_facet mds1 $LCTL set_param mdt.*.enable_chprojid_gid=-1
	$RUNAS lfs project -p $TSTPRJID $testdir/foo || error \
	"failed to set projid with normal user when enable_chprojid_gid=-1"

	$RUNAS lfs project -rC $testdir/ || error \
"failed to clear project state with normal user when enable_chprojid_gid=-1"

	touch $testdir/bar || error "failed touch $testdir/bar"
	$RUNAS lfs project -p $TSTPRJID $testdir/bar && error \
	"normal user should not be able to set projid on root owned file"

	change_project -p $TSTPRJID $testdir/bar || error \
		"root should be able to change its own file's projid"
}
run_test 66 "nonroot user can not change project state in default"

function cleanup_quota_test_75()
{
	do_facet mgs $LCTL nodemap_modify --name default \
		--property admin --value 1
	do_facet mgs $LCTL nodemap_modify --name default \
		--property trusted --value 1
	do_facet mgs $LCTL nodemap_modify --name default \
		--property squash_uid --value 99
	do_facet mgs $LCTL nodemap_modify --name default \
		--property squash_gid --value 99

	wait_nm_sync default admin_nodemap
	wait_nm_sync default trusted_nodemap

	do_facet mgs $LCTL nodemap_activate 0
	wait_nm_sync active

	resetquota -u $TSTUSR
}

test_dom_75() {
	local dd_failed=false
	local LIMIT=20480 #20M
	local qid=$TSTID

	for ((i = 0; i < $((LIMIT/2048-1)); i++)); do
		$DD of=$DIR/$tdir_dom/$tfile-$i count=1 \
			conv=fsync || dd_failed=true
	done

	$dd_failed && quota_error u $qid "write failed, expect succeed (1)"

	for ((i = $((LIMIT/2048-1)); i < $((LIMIT/1024 + 10)); i++)); do
		$DD of=$DIR/$tdir_dom/$tfile-$i count=1 \
			conv=fsync || dd_failed=true
	done

	$dd_failed || quota_error u $qid "write succeed, expect EDQUOT (1)"

	rm -f $DIR/$tdir_dom/*

	# flush cache, ensure noquota flag is set on client
	cancel_lru_locks
	sync; sync_all_data || true

	dd_failed=false

	$DD of=$DIR/$tdir/file count=$((LIMIT/2048-1)) conv=fsync ||
		quota_error u $qid "write failed, expect succeed (2)"

	for ((i = 0; i < $((LIMIT/2048 + 10)); i++)); do
		$DD of=$DIR/$tdir_dom/$tfile-$i count=1 \
			conv=fsync || dd_failed=true
	done

	$dd_failed || quota_error u $TSTID "write succeed, expect EDQUOT (2)"

	rm -f $DIR/$tdir/*
	rm -f $DIR/$tdir_dom/*

	# flush cache, ensure noquota flag is set on client
	cancel_lru_locks
	sync; sync_all_data || true

	dd_failed=false

	for ((i = 0; i < $((LIMIT/2048-1)); i++)); do
		$DD of=$DIR/$tdir_dom/$tfile-$i count=1 \
			conv=fsync || dd_failed=true
	done

	$dd_failed && quota_error u $qid "write failed, expect succeed (3)"

	$DD of=$DIR/$tdir/file count=$((LIMIT/2048 + 10)) oflag=direct &&
		quota_error u $qid "write succeed, expect EDQUOT (3)"
	true
}

test_75()
{
	local soft_limit=10 # MB
	local hard_limit=20 # MB
	local limit=$soft_limit
	local testfile="$DIR/$tdir/$tfile-0"
	local grace=20 # seconds
	local tdir_dom=${tdir}_dom

	if [ $(facet_fstype $SINGLEMDS) = "zfs" ]; then
	    grace=60
	fi

	(( $OST1_VERSION >= $(version_code 2.12.6.43) )) ||
		skip "need OST at least 2.12.6.43"

	setup_quota_test || error "setup quota failed with $?"
	stack_trap cleanup_quota_test_75 EXIT

	# enable ost quota
	set_ost_qtype $QTYPE || error "enable ost quota failed"
	set_mdt_qtype $QTYPE || error "enable mdt quota failed"

	local used=$(getquota -u $TSTID global curspace)
	$LFS setquota -t -u --block-grace $grace --inode-grace \
		$MAX_IQ_TIME $DIR || error "set user grace time failed"
	$LFS setquota -u $TSTUSR -b $((soft_limit+used/1024))M \
			-B $((hard_limit+used/1024))M -i 0 -I 0 $DIR ||
		error "set user quota failed"

	chmod 777 $DIR/$tdir || error "chmod 777 $DIR/$tdir failed"
	mkdir $DIR/$tdir_dom
	chmod 777 $DIR/$tdir_dom
	$LFS setstripe -E 1M -L mdt $DIR/$tdir_dom ||
		error "setstripe $tdir_dom failed"

	do_facet mgs $LCTL nodemap_activate 1
	wait_nm_sync active
	do_facet mgs $LCTL nodemap_modify --name default \
		--property admin --value 0
	do_facet mgs $LCTL nodemap_modify --name default \
		--property trusted --value 0
	do_facet mgs $LCTL nodemap_modify --name default \
		--property deny_unknown --value 0
	do_facet mgs $LCTL nodemap_modify --name default \
		--property squash_uid --value $TSTID
	do_facet mgs $LCTL nodemap_modify --name default \
		--property squash_gid --value $TSTID
	cancel_lru_locks mdc
	wait_nm_sync default admin_nodemap
	wait_nm_sync default trusted_nodemap
	wait_nm_sync default squash_uid

	# mmap write when over soft limit
	limit=$soft_limit
	$DD of=$testfile count=${limit} ||
		quota_error a  "root write failure, but expect success (1)"
	OFFSET=$((limit * 1024))
	cancel_lru_locks osc

	echo "Write to exceed soft limit"
	dd if=/dev/zero of=$testfile bs=1K count=10 seek=$OFFSET ||
	      quota_error a $TSTUSR "root write failure, but expect success (2)"
	OFFSET=$((OFFSET + 1024)) # make sure we don't write to same block
	cancel_lru_locks osc

	echo "mmap write when over soft limit"
	$MULTIOP $testfile.mmap OT40960SMW ||
		quota_error a $TSTUSR "mmap write failure, but expect success"
	cancel_lru_locks osc
	rm -f $testfile*
	wait_delete_completed || error "wait_delete_completed failed (1)"
	sync_all_data || true

	# test for user hard limit
	limit=$hard_limit
	log "Write..."
	$DD of=$testfile bs=1M count=$((limit/2)) ||
		quota_error u $TSTID \
			"root write failure, but expect success (3)"

	log "Write out of block quota ..."
	# possibly a cache write, ignore failure
	$DD of=$testfile bs=1M count=$((limit/2)) seek=$((limit/2)) || true
	# flush cache, ensure noquota flag is set on client
	cancel_lru_locks osc
	sync; sync_all_data || true
	# sync forced cache flush, but did not guarantee that slave
	# got new edquot through glimpse, so wait to make sure
	sleep 5
	$DD of=$testfile bs=1M count=1 seek=$limit conv=fsync &&
		quota_error u $TSTID \
			"user write success, but expect EDQUOT"
	rm -f $testfile
	wait_delete_completed || error "wait_delete_completed failed (2)"
	sync_all_data || true
	[ $(getquota -u $TSTUSR global curspace) -eq $used ] ||
		quota_error u $TSTID "user quota not released after deletion"

	test_dom_75
}
run_test 75 "nodemap squashed root respects quota enforcement"

test_76() {
	! is_project_quota_supported &&
		skip "skip project quota unsupported"

	setup_quota_test || error "setup quota failed with $?"
	quota_init

	local testfile="$DIR/$tdir/$tfile-0"

	touch $testfile
	$LFS project -p 4294967295 $testfile &&
		error "set project ID should fail"
	return 0
}
run_test 76 "project ID 4294967295 should be not allowed"

test_77()
{
	mount_client $MOUNT2 "ro"
	lfs setquota -u $TSTUSR -b 100M -B 100M -i 10K -I 10K $MOUNT2 &&
		error "lfs setquota should fail in read-only Lustre mount"
	umount $MOUNT2
}
run_test 77 "lfs setquota should fail in Lustre mount with 'ro'"

test_78()
{
	(( $OST1_VERSION >= $(version_code 2.12.8.6) )) ||
		skip "need OST at least 2.12.8.6"
	check_set_fallocate_or_skip

	setup_quota_test || error "setup quota failed with $?"

	# enable ost quota
	set_ost_qtype $QTYPE || error "enable ost quota failed"

	mkdir -p $DIR/$tdir || error "failed to create $tdir"
	chown $TSTUSR $DIR/$tdir || error "failed to chown $tdir"

	# setup quota limit
	$LFS setquota -u $TSTUSR -b25M -B25M $DIR/$tdir ||
		error "lfs setquota failed"

	# call fallocate
	runas -u $TSTUSR -g $TSTUSR fallocate -l 204800 $DIR/$tdir/$tfile

	kbytes=$(lfs quota -u $TSTUSR $DIR |
		awk -v pattern=$DIR 'match($0, pattern) {printf $2}')
	echo "kbytes returned:$kbytes"

	# For file size of 204800. We should be having roughly 200 kbytes
	# returned. Anything alarmingly low (50 taken as arbitrary value)
	# would bail out this TC. Also this also avoids $kbytes of 0
	# to be used in calculation below.
	(( $kbytes > 50 )) ||
		error "fallocate did not use quota. kbytes returned:$kbytes"

	local expect_lo=$(($kbytes * 95 / 100)) # 5% below
	local expect_hi=$(($kbytes * 105 / 100)) # 5% above

	# Verify kbytes is 200 (204800/1024). With a permited  5% drift
	(( $kbytes >= $expect_lo && $kbytes <= $expect_hi )) ||
		error "fallocate did not use quota correctly"
}
run_test 78 "Check fallocate increase quota usage"

test_78a()
{
	(( $CLIENT_VERSION >= $(version_code 2.12.9.28) )) ||
		skip "need client at least 2.12.9.28"
	(( $OST1_VERSION >= $(version_code 2.12.9.28) )) ||
		skip "need OST at least 2.12.9.28"
	check_set_fallocate_or_skip

	setup_quota_test || error "setup quota failed with $?"

	# enable ost quota
	set_ost_qtype $QTYPE || error "enable ost quota failed"

	mkdir -p $DIR/$tdir || error "failed to create $tdir"

	local projectid=5200 # Random project id to test

	change_project -sp $projectid $DIR/$tdir

	# setup quota limit
	$LFS setquota -p $projectid -b25M -B25M $DIR/$tdir ||
		error "lfs setquota project failed"

	# call fallocate
	fallocate -l 204800 $DIR/$tdir/$tfile

	# Get curspace (kbytes) for $projectid
	local kbytes=$(getquota -p $projectid global curspace)

	echo "kbytes returned:$kbytes"

	# For file size of 204800. We should be having roughly 200 kbytes
	# returned. Anything alarmingly low (50 taken as arbitrary value)
	# would bail out this TC. Also this also avoids $kbytes of 0
	# to be used in calculation below.
	(( $kbytes > 50 )) ||
		error "fallocate did not use projectid. kbytes returned:$kbytes"

	local expect_lo=$(($kbytes * 95 / 100)) # 5% below
	local expect_hi=$(($kbytes * 105 / 100)) # 5% above

	# Verify kbytes is 200 (204800/1024). With a permited  5% drift
	(( $kbytes >= $expect_lo && $kbytes <= $expect_hi )) ||
		error "fallocate did not use quota projectid correctly"
}
run_test 78a "Check fallocate increase projectid usage"

test_grace_with_default_quota()
{
	local qtype=$1
	local qdtype=$2
	local bgrace
	local igrace
	local bgrace2
	local igrace2
	echo "ttt1"
	$LFS setquota $qdtype -b 0 -B 0 -i 0 -I 0 $DIR ||
		error "clear default quota [$qdtype] failed"
	echo "ttt2"
	$LFS setquota -t $qtype --block-grace 1w --inode-grace 1w $DIR ||
		error "reset quota [$qdtype] grace failed"
	echo "ttt3"

	eval $($LFS quota -t $qtype $DIR | awk -F "[; ]" \
			'{printf("bgrace=%s;igrace=%s;", $4, $9)}')
	echo "ttt4"

	$LFS setquota $qdtype -B 10G -i 10k $DIR
	echo "ttt5"

	eval $($LFS quota -t $qtype $DIR | awk -F "[; ]" \
			'{printf("bgrace2=%s;igrace2=%s;", $4, $9)}')

	[ "$bgrace" == "$bgrace2" ] ||
			error "set default quota shouldn't affect block grace"
	[ "$igrace" == "$igrace2" ] ||
			error "set default quota shouldn't affect inode grace"

}

test_82()
{
	(( $MDS1_VERSION >= $(version_code 2.14.0.102) )) ||
		skip "need MDS 2.14.0.102 or later"
	is_project_quota_supported ||
		skip "skip project quota unsupported"

	setup_quota_test || error "setup quota failed with $?"
	stack_trap cleanup_quota_test
	quota_init

	local parent_dir="$DIR/$tdir.parent"
	local child_dir="$parent_dir/child"

	mkdir -p $child_dir
	stack_trap "chown -R 0:0 $parent_dir"

	chown $TSTUSR:$TSTUSR $parent_dir ||
		error "failed to chown on $parent_dir"
	chown $TSTUSR2:$TSTUSRS2 $child_dir ||
		error "failed to chown on $parent_dir"

	$LFS project -p 1000 $parent_dir ||
		error "failed to set project id on $parent_dir"
	$LFS project -p 1001 $child_dir ||
		error "failed to set project id on $child_dir"

	rmdir $child_dir || error "cannot remove child dir, test failed"
}
run_test 82 "verify more than 8 qids for single operation"

test_83()
{
	(( MDS1_VERSION >= $(version_code 2.12.8-ddn20) )) ||
		skip "Not supported before 2.12.8-ddn20."

	setup_quota_test || error "setup quota failed with $?"
	test_grace_with_default_quota "-u" "-U"
	test_grace_with_default_quota "-g" "-G"

	is_project_quota_supported || return 0
	test_grace_with_default_quota "-p" "-P"
}
run_test 83 "Setting default quota shouldn't affect grace time"

test_84()
{
	(( $MDS1_VERSION >= $(version_code 2.12.9) )) ||
		skip "need MDS 2.12.9 or later"

	local dir1="$DIR/$tdir/dir1"
	local TESTFILE1="$dir1/$tfile-1"
	local waited=0
	local grant=0

	setup_quota_test || error "setup quota failed with $?"
	quota_init
	set_ost_qtype $QTYPE || error "enable ost quota failed"

	$LFS setquota -g $TSTUSR -B 10G $DIR ||
		error "failed to set group quota for $TSTUSR"
	$LFS quota -gv $TSTUSR $DIR

	mkdir -p $dir1 || error "failed to mkdir"
	chown $TSTUSR.$TSTUSR $dir1 || error "chown $dir1 failed"

	$LFS setstripe -c 1 -i 0 $TESTFILE1
	$LFS getstripe $TESTFILE1
	chown $TSTUSR.$TSTUSR $TESTFILE1

	$RUNAS $DD of=$TESTFILE1 count=60 conv=nocreat oflag=direct ||
		quota_error g $TSTUSR "write failed"

	sync_all_data || true
	sleep 3
	$LFS quota -gv $TSTUSR $DIR

	# the grant quota should be larger than 0
	waited=0
	while (( $waited < 60 )); do
		grant=$(getquota -g $TSTUSR lustre-OST0000_UUID bhardlimit)
		(( ${grant} > 0 )) && break

		do_facet ost1 $LCTL set_param \
				osd-*.*-OST0000.quota_slave.force_reint=1
		sleep 1
		waited=$((waited + 1))
	done

	(( $waited >= 60)) && $LFS quota -gv $TSTUSR $DIR

	(( ${grant} > 0 )) || error "grant is not increased after dd"

#define OBD_FAIL_QUOTA_GRANT 0xA08
	lustre_fail mds 0xa08
	lustre_fail ost 0xa08
	sleep 1

	# clear quota limits to trigger updating grant quota
	$LFS setquota -g $TSTUSR -b 0 -B 0 $DIR ||
		error "failed to clear the group quota for $TSTUSR"
	$LFS quota -gv $TSTUSR $DIR

	# the grant quota should be set as insane value
	waited=0
	while (( $waited < 60 )); do
		grant=$(getquota -g $TSTUSR lustre-OST0000_UUID bhardlimit)
		(( ${#grant} == 20 )) && break

		sleep 1
		waited=$((waited + 1))
	done

	(( $waited >= 60)) && $LFS quota -gv $TSTUSR $DIR

	(( ${#grant} == 20 )) || error "grant is not set as insane value"

	lustre_fail mds_ost 0
	sleep 1

	# reset the quota
	$LFS setquota -g $TSTUSR -r $DIR ||
		error "failed to reset group quota for $TSTUSR"

	sleep 3
	$LFS quota -gv $TSTUSR $DIR

	# the grant quota should be reset
	grant=$(getquota -g $TSTUSR lustre-OST0000_UUID bhardlimit)
	(( ${#grant} == 20 )) && error "grant is not cleared"

	# test whether the quota still works
	$LFS setquota -g $TSTUSR -B 100M $DIR ||
		error "failed to set group quota for $TSTUSR"
	$LFS quota -gv $TSTUSR $DIR

	$RUNAS $DD of=$TESTFILE1 count=200 conv=nocreat oflag=direct &&
		quota_error g $TSTUSR "dd succeed, expect EDQUOT"

	$LFS setquota -g $TSTUSR -B 300M $DIR ||
		error "failed to set group quota for $TSTUSR"
	$LFS quota -gv $TSTUSR $DIR

	$RUNAS $DD of=$TESTFILE1 count=200 conv=nocreat oflag=direct ||
		quota_error g $TSTUSR "dd failed, expect succeed"
}
run_test 84 "Reset quota should fix the insane granted quota"

test_preacquired_quota()
{
	local test_dir=$1
	local qtype=$2
	local qtype_name=$3
	local qid=$4

	[[ "$qtype" == "-p" ]] && change_project -sp $qid $DIR/$tdir

	$LFS setquota $qtype $qid -i 100K -I 100K $DIR ||
		error "failed to set file [$qtype] quota"

	$RUNAS createmany -m $test_dir/tfile- 5000 ||
		error "failed to create files, expect succeed"

	wait_zfs_commit $SINGLEMDS
	$LFS setquota $qtype $qid -i 2K -I 2K $DIR ||
		error "failed to decrease file [$qtype] quota"

	wait_quota_synced "mds1" "MDT0000" $qtype_name $qid "hardlimit" "2048"

	# make sure the lqe->lqe_edquot is set
	$RUNAS createmany -m $test_dir/tfile2- 10
	sleep 5

	$RUNAS createmany -m $test_dir/tfile3- 30 &&
		error "succeed to create files, expect failed"

	rm -f $test_dir/tfile*
	$LFS setquota $qtype $qid -i 0 -I 0 $DIR ||
		error "failed to reset file user quota"
}

test_86()
{
	(( $MDS1_VERSION >= $(version_code 2.12.9) )) ||
		skip "need MDS >= 2.12.9 for quota over limit release fix"

	local test_dir="$DIR/$tdir/test_dir"

	setup_quota_test || error "setup quota failed with %?"
	set_mdt_qtype $QTYPE || error "enable mdt quota failed"

	$LFS setdirstripe -c 1 -i 0 $test_dir || error "setdirstripe failed"
	chmod 777 $test_dir

	test_preacquired_quota "$test_dir" "-u" "usr" "$TSTID"
	test_preacquired_quota "$test_dir" "-g" "grp" "$TSTID"

	is_project_quota_supported || return 0
	test_preacquired_quota "$test_dir" "-p" "prj" "1000"
}
run_test 86 "Pre-acquired quota should be released if quota is over limit"

quota_fini()
{
	do_nodes $(comma_list $(nodes_list)) "lctl set_param debug=-quota"
	disable_project_quota
}
reset_quota_settings
quota_fini

cd $ORIG_PWD
complete $SECONDS
check_and_cleanup_lustre
export QUOTA_AUTO=$QUOTA_AUTO_OLD
exit_status
